#!/usr/bin/env python

import numpy as np
from scipy.optimize import leastsq, least_squares
from scipy import integrate
import matplotlib.pyplot as plt
import sys
from intersect import intersection
import networkx as nx
import matplotlib as mpl
import json
from dicttoxml import dicttoxml
from xml.dom.minidom import parseString

# Setting up plotting class
plt.rcParams["mathtext.default"] = "regular"
plt.rcParams["font.family"] = "Arial"
plt.rc("font", size=22)  # controls default text sizes
plt.rc("axes", titlesize=22)  # fontsize of the axes title
plt.rc("axes", labelsize=22)  # fontsize of the x and y labels
plt.rc("xtick", labelsize=22)  # fontsize of the tick labels
plt.rc("ytick", labelsize=22)  # fontsize of the tick labels


class EOS:

    """EOS.

    This class contains methods to calculate the
    Equation of State for a selection of different
    models.

    """

    def __init__(self):
        """Initialize with obtaining fitting parameters
        for provided volume and energy data.
        If eostype=="energy" then it will take in data of
        energy and volume and differentiate for pressure.
        If eostype=="volume" then it will do similarly for
        pressure.
        If eostype=="phase_transition_pressure" it would
        plot the H vs P curves for each phase provided
        by the user as a list for infile.

        """

        return

    def _get_peos(self, dv=1e-3):
        """get_peos.

        This method uses a central difference method to
        differentiate EOS in energy to get pressure.

        central difference : (f(v+dv) - f(v-dv))/2dv

        P = dE/dV

        Parameters
        ----------
        v : float
            v
        dv : float
            dv

        """

        def P(e1, e2, dv):
            return -1 * (e2 - e1) / (2.00 * dv)

        if self.eostype != "enthalpy":

            v = self.vol_array

            v1 = v - dv
            v2 = v + dv

            # Murnaghan
            e1 = self.eos_murnaghan(self.eos_murnaghan_fitted.x, v1)
            e2 = self.eos_murnaghan(self.eos_murnaghan_fitted.x, v2)
            self.P_murnaghan = P(e1, e2, dv) * 1.60217e02  # GPa

            # Birch-Murnaghan
            e1 = self.eos_birch_murnaghan(self.eos_birch_murnaghan_fitted.x, v1)
            e2 = self.eos_birch_murnaghan(self.eos_birch_murnaghan_fitted.x, v2)
            self.P_birch_murnaghan = P(e1, e2, dv) * 1.60217e02  # GPa

            # Birch
            e1 = self.eos_birch(self.eos_birch_fitted.x, v1)
            e2 = self.eos_birch(self.eos_birch_fitted.x, v2)
            self.P_birch = P(e1, e2, dv) * 1.60217e02  # GPa

            # Vinet
            e1 = self.eos_vinet(self.eos_vinet_fitted.x, v1)
            e2 = self.eos_vinet(self.eos_vinet_fitted.x, v2)
            self.P_vinet = P(e1, e2, dv) * 1.60217e02  # GPa

        else:
            self.pressure = []
            for i, icoeffs in enumerate(self.selected_coeffs):

                v = self.vol_array[i]

                v1 = v - dv
                v2 = v + dv

                if self.model[i] == "Vinet":
                    e1 = self.eos_vinet(icoeffs, v1)
                    e2 = self.eos_vinet(icoeffs, v2)
                    self.pressure.append(P(e1, e2, dv))

                elif self.model[i] == "Birch":
                    e1 = self.eos_birch(icoeffs, v1)
                    e2 = self.eos_birch(icoeffs, v2)
                    self.pressure.append(P(e1, e2, dv))

                elif self.model[i] == "Murnaghan":
                    e1 = self.eos_murnaghan(icoeffs, v1)
                    e2 = self.eos_murnaghan(icoeffs, v2)
                    self.pressure.append(P(e1, e2, dv))

                elif self.model[i] == "Birch-Murnaghan":
                    e1 = self.eos_birch_murnaghan(icoeffs, v1)
                    e2 = self.eos_birch_murnaghan(icoeffs, v2)
                    self.pressure.append(P(e1, e2, dv))

            # Converting self.pressure into numpy array and converting to GPa
            self.pressure = 1.60217e02 * np.array(self.pressure, dtype="float64")

    def _get_eeos(self):
        """get_eeos.

        This method uses scipy.integrate to integrate
        pressure to obtain energy.


        """

        v = self.volume
        P_birch_murnaghan = self.eos_birch_murnaghan_pressure(
            self.eos_birch_murnaghan_pressure_fitted.x, v
        )

        self.E_birch_murnaghan = (
            integrate.cumtrapz(v, P_birch_murnaghan, initial=0) * 6.2415e-03
        )  # eV

    def plot_eos(
        self,
        infile=None,
        eostype="energy",
        natoms=1,
        au=False,
        vlim=None,
        model=None,
        raw_data=True,
        export=False,
    ):
        """plot_eos.

        This funciton plots the fitted EOS models.
        Initialize with obtaining fitting parameters
        for provided volume and energy data.
        If eostype=="energy" then it will take in data of
        energy and volume and differentiate for pressure.
        If eostype=="volume" then it will do similarly for
        pressure.

        Parameters
        ----------
            infile : list, optional (default ``None``)
                List of input files.

            eostype : str, optional (default ``"energy"``)
                Type of EOS input data; energy or pressure.

            natoms : int, optional (default ``1``)
                Number of atoms

            au : bool, optional (default ``False``)
                Output in atomic units.

            vlim : list, optional (default ``None``)
               Minimum and maximum for volume range.

            model : str, optional (default ``None``)
                Name of model; Vinet, Birch, Murnaghan, Birch-Murnaghan

            raw_data : bool, optional (default ``True``)
                Also include the raw input data in the plot.

            export : bool, optional (default ``False``)
                Export data as .json and .xml.

        Returns
        -------
            param : None
        """

        self.eostype = eostype
        self.natoms = natoms
        self.au = au
        self.vlim = vlim
        self.model = model
        self.raw_data = raw_data
        self.export = export

        if self.eostype == "energy":

            # reading the files with volume and energy
            # columns.

            fi = open(infile, "r")
            data = fi.readlines()
            fi.close()

            self.volume = []
            self.energy = []

            for i in range(len(data)):
                self.volume.append(float(data[i].split()[0]))
                self.energy.append(float(data[i].split()[1]))

            self.volume = np.array(self.volume)
            self.energy = np.array(self.energy)

            a, b, c = np.polyfit(self.volume, self.energy, 2)

            V0 = -b / (2 * a)
            E0 = a * V0 ** 2 + b * V0 + c
            B0 = 2 * a * V0
            Bp = 4.0

            self.coeffs0 = [E0, B0, Bp, V0]

        elif self.eostype == "pressure":

            # reading the files with volume and energy
            # columns.

            fi = open(infile, "r")
            data = fi.readlines()
            fi.close()

            self.volume = []
            self.pressure = []

            for i in range(len(data)):
                self.volume.append(float(data[i].split()[0]))
                self.pressure.append(float(data[i].split()[1]))

            self.volume = np.array(self.volume)
            self.pressure = np.array(self.pressure)

            a, b, c = np.polyfit(self.volume, self.pressure, 2)

            V0 = -b / (2 * a)
            P0 = a * V0 ** 2 + b * V0 + c
            K0 = 2 * a * V0
            Kp = 4.0

            self.coeffs0 = [P0, K0, Kp, V0]

        # Interpolating volume array
        if self.vlim:
            self.vol_array = np.linspace(self.vlim[0], self.vlim[1], 1000)
        else:
            self.vol_array = np.linspace(np.min(self.volume), np.max(self.volume), 1000)

        # Call fitting function
        self.fitting()

        if self.eostype == "energy":
            # Call Pressure calculation function
            self._get_peos()
        elif self.eostype == "pressure":
            # Call Energy calculation function
            self._get_eeos()

        fig = plt.figure(figsize=(13, 9))
        axs = fig.add_subplot(111)

        # eV to Ha and A^3 to Bohr^3 conversion if au=True
        if self.au:
            au_converter_e = 0.0367493
            au_converter_v = 6.748376012323316
        else:
            au_converter_e = 1.0
            au_converter_v = 1.0

        if self.eostype == "energy":

            # Energy vs Volume plot

            if self.model is None:

                axs.plot(
                    au_converter_v * self.vol_array / self.natoms,
                    au_converter_e
                    * self.eos_vinet(self.eos_vinet_fitted.x, self.vol_array)
                    / self.natoms,
                    "orangered",
                    linewidth=11,
                    label="Vinet",
                )

                axs.plot(
                    au_converter_v * self.vol_array / self.natoms,
                    au_converter_e
                    * self.eos_birch(self.eos_birch_fitted.x, self.vol_array)
                    / self.natoms,
                    "lime",
                    linestyle="solid",
                    linewidth=7,
                    label="Birch",
                )

                axs.plot(
                    au_converter_v * self.vol_array / self.natoms,
                    au_converter_e
                    * self.eos_murnaghan(self.eos_murnaghan_fitted.x, self.vol_array)
                    / self.natoms,
                    color="magenta",
                    linestyle="dashed",
                    linewidth=5,
                    label="Murnaghan",
                )

                axs.plot(
                    au_converter_v * self.vol_array / self.natoms,
                    au_converter_e
                    * self.eos_birch_murnaghan(
                        self.eos_birch_murnaghan_fitted.x, self.vol_array
                    )
                    / self.natoms,
                    "k--",
                    linewidth=2,
                    label="Birch-Murnaghan",
                )

                # axs.scatter(
                #     au_converter_v * self.volume,
                #     au_converter_e * self.energy / self.natoms,
                #     s=600,
                #     facecolors="none",
                #     edgecolors="black",
                #     label="Raw Data",
                # )

            elif self.model == "Vinet":
                axs.plot(
                    au_converter_v * self.vol_array / self.natoms,
                    au_converter_e
                    * self.eos_vinet(self.eos_vinet_fitted.x, self.vol_array)
                    / self.natoms,
                    "r",
                    linewidth=2,
                    label="Vinet",
                )

            elif self.model == "Birch":

                axs.plot(
                    au_converter_v * self.vol_array / self.natoms,
                    au_converter_e
                    * self.eos_birch(self.eos_birch_fitted.x, self.vol_array)
                    / self.natoms,
                    "r",
                    linewidth=2,
                    label="Birch",
                )

            elif self.model == "Murnaghan":
                axs.plot(
                    au_converter_v * self.vol_array / self.natoms,
                    au_converter_e
                    * self.eos_murnaghan(self.eos_murnaghan_fitted.x, self.vol_array)
                    / self.natoms,
                    color="r",
                    linewidth=2,
                    label="Murnaghan",
                )

            elif self.model == "Birch-Murnaghan":
                axs.plot(
                    au_converter_v * self.vol_array / self.natoms,
                    au_converter_e
                    * self.eos_birch_murnaghan(
                        self.eos_birch_murnaghan_fitted.x, self.vol_array
                    )
                    / self.natoms,
                    "r",
                    linewidth=2,
                    label="Birch-Murnaghan",
                )

            if self.raw_data:
                axs.scatter(
                    au_converter_v * self.volume / self.natoms,
                    au_converter_e * self.energy / self.natoms,
                    s=600,
                    facecolors="none",
                    edgecolors="black",
                    label="Raw Data",
                )

            if self.au:
                axs.set_xlabel("Volume ($Bohr^3$/atom)")
                axs.set_ylabel("Energy (Ha/atom)")
            else:
                axs.set_xlabel("Volume ($\AA^3$/atom)")
                axs.set_ylabel("Energy (eV/atom)")
            axs.set_title("Energy vs. Volume")
            plt.legend(loc="best")
            plt.show()

            # Pressure vs Volume

            fig = plt.figure(figsize=(13, 9))
            axs2 = fig.add_subplot(111)

            if self.model is None:

                axs2.plot(
                    au_converter_v * self.vol_array / self.natoms,
                    self.P_vinet,
                    "orangered",
                    linewidth=11,
                    label="Vinet",
                )

                axs2.plot(
                    au_converter_v * self.vol_array / self.natoms,
                    self.P_birch,
                    "lime",
                    linestyle="solid",
                    linewidth=7,
                    label="Birch",
                )

                axs2.plot(
                    au_converter_v * self.vol_array / self.natoms,
                    self.P_murnaghan,
                    color="magenta",
                    linestyle="dashed",
                    linewidth=5,
                    label="Murnaghan",
                )

                axs2.plot(
                    au_converter_v * self.vol_array / self.natoms,
                    self.P_birch_murnaghan,
                    "k--",
                    linewidth=2,
                    label="Birch-Murnaghan",
                )

            elif self.model == "Vinet":
                axs2.plot(
                    au_converter_v * self.vol_array / self.natoms,
                    self.P_vinet,
                    "r",
                    linewidth=2,
                    label="Vinet",
                )

            elif self.model == "Birch":
                axs2.plot(
                    au_converter_v * self.vol_array / self.natoms,
                    self.P_birch,
                    "r",
                    linewidth=2,
                    label="Birch",
                )

            elif self.model == "Murnaghan":
                axs2.plot(
                    au_converter_v * self.vol_array / self.natoms,
                    self.P_murnaghan,
                    color="r",
                    linewidth=2,
                    label="Murnaghan",
                )

            elif self.model == "Birch-Murnaghan":
                axs2.plot(
                    au_converter_v * self.vol_array / self.natoms,
                    self.P_birch_murnaghan,
                    "r",
                    linewidth=2,
                    label="Birch-Murnaghan",
                )

            if self.au:
                axs2.set_xlabel("Volume ($Bohr^3$/atom)")
            else:
                axs2.set_xlabel("Volume ($\AA^3$/atom)")
            axs2.set_ylabel("Pressure (GPa)")
            axs2.set_title("Pressure vs. Volume")
            plt.legend(loc="best")
            plt.show()

        elif self.eostype == "pressure":
            # Pressure vs Volume plot

            ## TODO: Add other models

            if self.model is None:

                axs.plot(
                    au_converter_v * self.vol_array / self.natoms,
                    self.eos_birch_murnaghan_pressure(
                        self.eos_birch_murnaghan_pressure_fitted.x, self.vol_array
                    ),
                    "r-",
                    linewidth=2,
                    label="Birch-Murnaghan",
                )

                # axs.scatter(
                #     au_converter_v * self.volume,
                #     self.pressure / self.natoms,
                #     s=600,
                #     facecolors="none",
                #     edgecolors="black",
                #     label="Raw Data",
                # )

            elif self.model == "Birch-Murnaghan":
                axs.plot(
                    au_converter_v * self.vol_array / self.natoms,
                    self.eos_birch_murnaghan_pressure(
                        self.eos_birch_murnaghan_pressure_fitted.x, self.vol_array
                    ),
                    "r-",
                    linewidth=2,
                    label="Birch-Murnaghan",
                )

            if self.raw_data:
                axs.scatter(
                    au_converter_v * self.volume / self.natoms,
                    self.pressure / self.natoms,
                    s=600,
                    facecolors="none",
                    edgecolors="black",
                    label="Raw Data",
                )

            if self.au:
                axs.set_xlabel("Volume ($Bohr^3$/atom)")
            else:
                axs.set_xlabel("Volume ($\AA^3$/atom)")
            axs.set_ylabel("Pressure (GPa)")
            axs.set_title("Pressure vs. Volume")
            plt.legend(loc="best")
            plt.show()

            # Energy vs Volume

            fig = plt.figure(figsize=(13, 9))
            axs2 = fig.add_subplot(111)

            if self.model is None:

                axs2.plot(
                    au_converter_v * self.volume / self.natoms,
                    au_converter_e * self.E_birch_murnaghan / self.natoms,
                    "r-",
                    linewidth=2,
                    label="Birch-Murnaghan",
                )

            elif self.model == "Birch-Murnaghan":
                axs2.plot(
                    au_converter_v * self.volume / self.natoms,
                    au_converter_e * self.E_birch_murnaghan / self.natoms,
                    "r-",
                    linewidth=2,
                    label="Birch-Murnaghan",
                )

            if self.au:
                axs2.set_ylabel("Energy (Ha/atom)")
                axs2.set_xlabel("Volume ($Bohr^3$/atom)")
            else:
                axs2.set_ylabel("Energy (eV/atom)")
                axs2.set_xlabel("Volume ($\AA^3$/atom)")
            axs2.set_title("Energy vs. Volume")
            plt.legend(loc="best")
            plt.show()

        # Saving as .json and .xml
        if self.export:
            self.to_json()
            self.to_xml()

    def plot_enthalpy_curves(
        self,
        infiles=None,
        natoms=1,
        au=False,
        vlim=None,
        vlim_list=None,
        deltaH_index=None,
        labels=None,
        export=False,
    ):
        """plot_enthalpy_curves.

        Initialize with obtaining fitting parameters
        for provided volume and energy data.
        This function plots Enthalpy vs Pressure curves
        for multiple datasets and calculates the phase transition pressure.
        Additionally it plots enthalpy differences with respect to a selected
        phase.


        Parameters
        ----------
            infiles : list, optional (default ``None``)
                List of input files.

            natoms : int, optional (default ``1``)
                Number of atoms.

            labels : str, optional (default ``None``)
                List of names of phases. Supports Latex.
                If not provided infiles will be used.

            au : bool, optional (default ``False``)
                Output in atomic units?

            vlim : list, optional (default ``None``)
                Minimum and maximum volume.

            vlim_list : list, optional (default ``None``)
                List of minimum and maximum volumes for each file.

            deltaH_index : int, optional (default ``None``)
                Index of file to calculate enthalpy difference with respect to.

            export : bool, optional (default ``False``)
                Export data as .json and .xml.

        Returns
        -------
            param : None
        """

        self.eostype = "enthalpy"
        self.infiles = infiles
        self.natoms = natoms
        self.au = au
        self.vlim = vlim
        self.vlim_list = vlim_list
        self.deltaH_index = deltaH_index
        self.labels = labels
        self.export = export

        nfiles = len(self.infiles)
        self.volume = []
        self.energy = []
        self.vol_array = []

        # read a list of files for different phases.
        for ifile, fname in enumerate(self.infiles):
            volume_temp = []
            energy_temp = []
            with open(fname) as fi:
                rawdata = fi.readlines()
                for i in range(len(rawdata)):
                    volume_temp.append(float(rawdata[i].split()[0]))
                    energy_temp.append(float(rawdata[i].split()[1]))

            self.volume.append(volume_temp)
            self.energy.append(energy_temp)

        # Converting 2D lists to numpy arrays
        self.volume = np.array(self.volume, dtype="float64")
        self.energy = np.array(self.energy, dtype="float64")

        # initial fitting coefficients
        self.coeffs0 = []
        for i in range(nfiles):
            a, b, c = np.polyfit(self.volume[i], self.energy[i], 2)

            V0 = -b / (2 * a)
            E0 = a * V0 ** 2 + b * V0 + c
            B0 = 2 * a * V0
            Bp = 4.0

            self.coeffs0.append([E0, B0, Bp, V0])

        # Converting to a numpy array
        self.coeffs0 = np.array(self.coeffs0, dtype="float64")

        # Interpolating volume arrays and storing in a 2D self.vol_array
        for i in range(nfiles):
            if self.vlim:
                self.vol_array.append(np.linspace(self.vlim[0], self.vlim[1], 2000))
            elif self.vlim_list:
                self.vol_array.append(
                    np.linspace(self.vlim_list[i][0], self.vlim_list[i][1], 2000)
                )
            else:
                self.vol_array.append(
                    np.linspace(np.min(self.volume[i]), np.max(self.volume[i]), 2000)
                )

        # converting self.vol_array to numpy array
        self.vol_array = np.array(self.vol_array, dtype="float64")

        # Call fitting function
        self.fitting()

        # Call Pressure calculation function
        self._get_peos()

        # Calculating enthalpy for each phase
        # Converting PV from GPaA3 -> eV

        self.H = []
        for i in range(nfiles):

            if self.model[i] == "Vinet":
                self.H.append(
                    self.eos_vinet(self.selected_coeffs[i], self.vol_array[i])
                    + np.multiply(self.pressure[i], self.vol_array[i])
                    * (1 / 160.2176621)
                )

            elif self.model[i] == "Birch":
                self.H.append(
                    self.eos_birch(self.selected_coeffs[i], self.vol_array[i])
                    + np.multiply(self.pressure[i], self.vol_array[i])
                    * (1 / 160.2176621)
                )
            elif self.model[i] == "Murnaghan":
                self.H.append(
                    self.eos_murnaghan(self.selected_coeffs[i], self.vol_array[i])
                    + np.multiply(self.pressure[i], self.vol_array[i])
                    * (1 / 160.2176621)
                )
            elif self.model[i] == "Birch-Murnaghan":
                self.H.append(
                    self.eos_birch_murnaghan(self.selected_coeffs[i], self.vol_array[i])
                    + np.multiply(self.pressure[i], self.vol_array[i])
                    * (1 / 160.2176621)
                )

        # Converting self.H to numpy array
        self.H = np.array(self.H, dtype="float64")

        # Normalize H for number of atoms
        for i, iatom in enumerate(self.natoms):
            self.H[i] = self.H[i] / iatom

        # Convert eV to Ha if au=True
        if self.au:
            self.H = 0.0367493 * self.H

        # Enthalpy vs Pressure plots

        fig = plt.figure(figsize=(13, 9))
        axs = fig.add_subplot(111)

        # Matrix to store transition information
        self.trans_mat = []

        # Intersection points
        if nfiles > 1:
            # print("\nEquilibrium enthalpy points between phases : ")
            for i in range(nfiles):
                for j in range(nfiles):
                    if i < j:
                        # print("\n%s and %s:" % (self.infiles[i], self.infiles[j]))
                        self.int_x, self.int_y = intersection(
                            self.pressure[i], self.H[i], self.pressure[j], self.H[j]
                        )

                        # Check if there are multiple intersection points
                        if self.int_x.size:
                            if self.int_x.size < 2:
                                if self.au:
                                    # print(
                                    #     "%.2f GPa at %.2f Ha "
                                    #     % (self.int_x, self.int_y)
                                    # )
                                    self.trans_mat.append(
                                        [
                                            self.infiles[i],
                                            self.infiles[j],
                                            self.int_x,
                                            self.int_y,
                                        ]
                                    )
                                else:
                                    # print(
                                    #     "%.2f GPa at %.2f eV "
                                    #     % (self.int_x, self.int_y)
                                    # )
                                    self.trans_mat.append(
                                        [
                                            self.infiles[i],
                                            self.infiles[j],
                                            self.int_x,
                                            self.int_y,
                                        ]
                                    )

                            else:
                                for jj in range(self.int_x.size):
                                    if self.au:
                                        # print(
                                        #     "%.2f GPa at %.2f Ha "
                                        #     % (self.int_x[jj], self.int_y[jj])
                                        # )
                                        self.trans_mat.append(
                                            [
                                                self.infiles[i],
                                                self.infiles[j],
                                                self.int_x,
                                                self.int_y,
                                            ]
                                        )

                                    else:
                                        # print(
                                        #     "%.2f GPa at %.2f eV "
                                        #     % (self.int_x[jj], self.int_y[jj])
                                        # )
                                        self.trans_mat.append(
                                            [
                                                self.infiles[i],
                                                self.infiles[j],
                                                self.int_x,
                                                self.int_y,
                                            ]
                                        )

                        # else:
                        #     print("NOT FOUND FOR THIS VOLUME RANGE!")
                        # # plt.plot(x, y, "*k")
        self.trans_mat = np.array(self.trans_mat, dtype="object")

        # reordered trans_mat to contain minimum enthalpy values for multiple
        # intersections and their corresponding pressure.
        self.trans_mat_reordered = np.zeros((self.trans_mat.shape), dtype="object")
        self.trans_mat_reordered[:, 0:2] = self.trans_mat[:, 0:2]
        self.trans_mat_reordered[:, 3] = [
            min(self.trans_mat[i, 3]).item() for i in range(len(self.trans_mat))
        ]
        indx_arr = [np.argmin(self.trans_mat[i, 3]) for i in range(len(self.trans_mat))]
        self.trans_mat_reordered[:, 2] = [
            self.trans_mat[i, 2][j] for i, j in enumerate(indx_arr)
        ]
        self.trans_mat_reordered = self.trans_mat_reordered[
            np.argsort(self.trans_mat_reordered[:, 3])
        ]
        self.trans_mat_reordered_tmp = self.trans_mat_reordered.copy()

        # Find lowest curve of the two intersecting curves for each
        # row in trans_mat_reordered. c1, c2 are the indexes of the pressure
        # arrays.
        for i in range(len(self.trans_mat_reordered)):
            c1 = np.where(
                np.isclose(
                    self.pressure[self.infiles.index(self.trans_mat_reordered[i][0])],
                    self.trans_mat_reordered[i][2],
                    1e-03,
                )
            )
            c2 = np.where(
                np.isclose(
                    self.pressure[self.infiles.index(self.trans_mat_reordered[i][1])],
                    self.trans_mat_reordered[i][2],
                    1e-03,
                )
            )

            # corresponding enthalpy values for the two intersecting curves
            # Adding 500 to see a point a bit down along the interseciton point.
            h1 = self.H[self.infiles.index(self.trans_mat_reordered[i][0])][
                c1[0][-1] + 500
            ]
            h2 = self.H[self.infiles.index(self.trans_mat_reordered[i][1])][
                c1[0][-1] + 500
            ]

            # reorder the phases such that lowest energy phase appears first
            if h1 > h2:
                self.trans_mat_reordered_tmp[i][0] = self.trans_mat_reordered[i][1]
                self.trans_mat_reordered_tmp[i][1] = self.trans_mat_reordered[i][0]
            else:
                self.trans_mat_reordered_tmp[i][0] = self.trans_mat_reordered[i][0]
                self.trans_mat_reordered_tmp[i][1] = self.trans_mat_reordered[i][1]

        # setting to original array
        self.trans_mat_reordered = self.trans_mat_reordered_tmp
        # Adding additional phase for testing
        # tm = np.append(self.trans_mat_reordered, ["Im-3m", "XXX", 8.0, -2.0])
        # self.trans_mat_reordered = tm.reshape(7, 4)

        print("\nPossible transition paths for the provided volume range:\n")
        print(
            "NOTE: Ordered with ascending enthalpy. i.e. topmost transition is the most probable.\n"
        )
        selected_array = []
        selected_array2 = []
        for i in range(len(self.trans_mat_reordered)):
            for j in range(len(self.trans_mat_reordered)):
                if i < j:
                    if self.trans_mat_reordered[i][1] == self.trans_mat_reordered[j][0]:
                        selected_array.append(i)

                        print(
                            "{} -> {} (at {:0.3f} GPa) -> {} (at {:0.3f} GPa)".format(
                                self.trans_mat_reordered[i][0],
                                self.trans_mat_reordered[i][1],
                                self.trans_mat_reordered[i][2],
                                self.trans_mat_reordered[j][1],
                                self.trans_mat_reordered[j][2],
                            ),
                        )

                        for k in range(len(self.trans_mat_reordered)):
                            if j < k:
                                if (
                                    self.trans_mat_reordered[j][1]
                                    == self.trans_mat_reordered[k][0]
                                ):
                                    selected_array2.append(j)

                        if j in selected_array2:
                            print(
                                "-> {} (at {:0.3f} GPa)".format(
                                    self.trans_mat_reordered[k][1],
                                    self.trans_mat_reordered[k][2],
                                ),
                            )

            if i not in selected_array:
                print(
                    "{} -> {} (at {:0.3f} GPa)".format(
                        self.trans_mat_reordered[i][0],
                        self.trans_mat_reordered[i][1],
                        self.trans_mat_reordered[i][2],
                    ),
                )

        # print("\nPossible transition paths for the provided volume range:\n")
        # selected_array = []
        # selected_array2 = []
        # for i in range(len(self.trans_mat_reordered)):
        #     for j in range(len(self.trans_mat_reordered)):

        #         def recursion(i, j):
        #             if i < j:
        #                 if (
        #                     self.trans_mat_reordered[i][1]
        #                     == self.trans_mat_reordered[j][0]
        #                 ):
        #                     selected_array.append([])
        #                     selected_array.append(i)

        #                     print(
        #                         "{} -> {} -> {}".format(
        #                             self.trans_mat_reordered[i][0],
        #                             self.trans_mat_reordered[i][1],
        #                             self.trans_mat_reordered[j][1],
        #                         ),
        #                         end="\n",
        #                     )

        #                     for k in range(len(self.trans_mat_reordered)):
        #                         if j < k:
        #                             if (
        #                                 self.trans_mat_reordered[j][1]
        #                                 == self.trans_mat_reordered[k][0]
        #                             ):
        #                                 selected_array2.append(j)

        #                     if j in selected_array:
        #                         print(
        #                             "-> {}".format(self.trans_mat_reordered[k][1]),
        #                         )

        #     if i not in selected_array:
        #         print(
        #             "{} -> {}".format(
        #                 self.trans_mat_reordered[i][0], self.trans_mat_reordered[i][1]
        #             ),
        #         )

        if self.labels is None:
            self.labels = self.infiles

        for i, filename in enumerate(self.infiles):
            axs.plot(self.pressure[i], self.H[i], label=self.labels[i])
        axs.set_xlabel("Pressure (GPa)")
        if self.au:
            axs.set_ylabel("Enthalpy (Ha/atom)")
        else:
            axs.set_ylabel("Enthalpy (eV/atom)")
        axs.set_title("Enthalpy vs. Pressure")
        plt.legend(loc="best")
        axs.grid(color="gainsboro", ls="--", lw=0.6)
        plt.savefig("enthalpy.pdf")
        plt.show()

        # Network map for phase transitions
        # Added by Pedram
        cmap = plt.get_cmap("nipy_spectral")
        DG = nx.DiGraph()
        names = np.unique(self.trans_mat_reordered[:, 0:2])

        DG.add_nodes_from(names)
        for trans in self.trans_mat_reordered:
            DG.add_edge(
                trans[0],
                trans[1],
                weight=trans[3],
            )
        pos = nx.layout.spring_layout(DG)

        node_sizes = [3500] * DG.number_of_nodes()
        edge_colors = [
            DG.get_edge_data(edge[0], edge[1])["weight"] for edge in DG.edges
        ]  # self.tras_mat_reordered[:, 3]

        plt.figure(figsize=(13, 9))
        nodes = nx.draw_networkx_nodes(
            DG, pos, node_size=node_sizes, node_color="gainsboro"
        )
        edges = nx.draw_networkx_edges(
            DG,
            pos,
            node_size=node_sizes,
            # arrowstyle="->",
            arrowsize=16,
            edge_color=edge_colors,
            edge_cmap=cmap,
            width=4,
        )

        nx.draw_networkx_labels(DG, pos, font_size=16)

        pc = mpl.collections.PatchCollection(edges, cmap=cmap)
        pc.set_array(edge_colors)
        cb = plt.colorbar(pc, orientation="vertical")
        if self.au:
            cb.set_label("Enthalpy (Ha/atom)")
        else:
            cb.set_label("Enthalpy (eV/atom)")
        ax = plt.gca()
        ax.set_axis_off()
        plt.savefig("enthalpy-network.pdf")
        plt.show()

        # Enthalpy differences with respect to a selected phase
        if self.deltaH_index:

            # Converting human input index to python 0 index
            self.deltaH_index = int(self.deltaH_index)

            # name of phase
            phase_name = self.infiles[self.deltaH_index]

            # removing selected phase from infiles list
            # self.infiles.remove(phase_name)

            # max and min pressure of selected phase
            self.plim = [
                min(self.pressure[self.deltaH_index]),
                max(self.pressure[self.deltaH_index]),
            ]

            # Pressure matrix with removed row
            # self.pressure = np.delete(self.pressure, self.deltaH_index, axis=0)

            # Enthalpy matrix with removed row
            # self.deltaH = np.delete(
            #     (self.H - self.H[self.deltaH_index]), self.deltaH_index, axis=0
            # )

            # Enthalpy difference
            self.deltaH = self.H - self.H[self.deltaH_index]

            fig = plt.figure(figsize=(13, 9))
            axs2 = fig.add_subplot(111)

            for i, filename in enumerate(self.infiles):
                axs2.plot(self.pressure[i], self.deltaH[i], label=self.labels[i])
            axs2.set_xlabel("Pressure (GPa)")
            if self.au:
                axs2.set_ylabel("$\Delta$H (Ha/atom)")
            else:
                axs2.set_ylabel("$\Delta$H (eV/atom)")
            title_string = "$\Delta$H vs. Pressure wrt " + str(phase_name)
            axs2.set_title(title_string)
            # axs2.axhline(color="black")
            axs2.set_xlim(self.plim)
            plt.legend(loc="best")
            plt.show()

        # Saving as .json and .xml
        if self.export:
            self.to_json()
            self.to_xml()

    ############################# FITTING #############################

    def fitting(self):
        """fitting.

        Fits coefficients using the least square method.
        returns an array of fitted coefficients [E0,B0,Bp,V0] or
        [P0,K0,Kp,V0].

        """

        if self.eostype == "energy":

            print(
                "Fitting coefficients [E0 (eV), B0 (GPa), Bp (GPa), V0 (Anstroms^3)]\nalong with Mean Squared Error (MSE)\n"
            )
            np.set_printoptions(formatter={"float": "{: 0.3f}".format})

            # Vinet
            def residuals_vinet(coeffs, y, x):
                return y - self.eos_vinet(coeffs, x)

            self.eos_vinet_fitted = least_squares(
                residuals_vinet, self.coeffs0, args=(self.energy, self.volume)
            )
            print("Vinet : %s" % self.eos_vinet_fitted.x)
            print(
                "Vinet (MSE) : {:.3e}\n".format(np.mean(self.eos_vinet_fitted.fun ** 2))
            )

            # Birch
            def residuals_birch(coeffs, y, x):
                return y - self.eos_birch(coeffs, x)

            self.eos_birch_fitted = least_squares(
                residuals_birch, self.coeffs0, args=(self.energy, self.volume)
            )
            print("Birch : %s" % self.eos_birch_fitted.x)
            print(
                "Birch (MSE) : {:.3e}\n".format(np.mean(self.eos_birch_fitted.fun ** 2))
            )

            # Murnaghan
            def residuals_murnaghan(coeffs, y, x):
                return y - self.eos_murnaghan(coeffs, x)

            self.eos_murnaghan_fitted = least_squares(
                residuals_murnaghan, self.coeffs0, args=(self.energy, self.volume)
            )

            print("Murnaghan : %s" % self.eos_murnaghan_fitted.x)
            print(
                "Murnaghan (MSE) : {:.3e}\n".format(
                    np.mean(self.eos_murnaghan_fitted.fun ** 2)
                )
            )

            # Birch-Murnaghan
            def residuals_birch_murnaghan(coeffs, y, x):
                return y - self.eos_birch_murnaghan(coeffs, x)

            self.eos_birch_murnaghan_fitted = least_squares(
                residuals_birch_murnaghan, self.coeffs0, args=(self.energy, self.volume)
            )
            print("Birch-Murnaghan : %s" % self.eos_birch_murnaghan_fitted.x)
            print(
                "Birch-Murnaghan (MSE) : {:.3e}\n".format(
                    np.mean(self.eos_birch_murnaghan_fitted.fun ** 2)
                )
            )

            cost_array = [
                np.mean(self.eos_vinet_fitted.fun ** 2),
                np.mean(self.eos_birch_fitted.fun ** 2),
                np.mean(self.eos_murnaghan_fitted.fun ** 2),
                np.mean(self.eos_birch_murnaghan_fitted.fun ** 2),
            ]
            eos_name = ["Vinet", "Birch", "Murnaghan", "Birch-Murnaghan"]

            print(
                "Based on MSE %s is the best EOS model for this dataset."
                % (eos_name[cost_array.index(min(cost_array))])
            )

        elif self.eostype == "pressure":
            print(
                "Fitting coefficients [P0 (GPa), K0 (GPa), Kp (GPa), V0 (Anstroms^3)]\nalong with Mean Squared Error (MSE)\n"
            )
            np.set_printoptions(formatter={"float": "{: 0.3f}".format})

            # Birch-Murnaghan Pressure
            def residuals_birch_murnaghan_pressure(coeffs, y, x):
                return y - self.eos_birch_murnaghan_pressure(coeffs, x)

            self.eos_birch_murnaghan_pressure_fitted = least_squares(
                residuals_birch_murnaghan_pressure,
                self.coeffs0,
                args=(self.pressure, self.volume),
            )
            print("Birch-Murnaghan : %s" % self.eos_birch_murnaghan_pressure_fitted.x)
            print(
                "Birch-Murnaghan (MSE) : {:.3e}\n".format(
                    np.mean(self.eos_birch_murnaghan_pressure_fitted.fun ** 2)
                )
            )

            cost_array = [
                np.mean(self.eos_birch_murnaghan_pressure_fitted.fun ** 2),
            ]
            eos_name = ["Birch-Murnaghan"]

            print(
                "Based on MSE %s is the best EOS model for this dataset."
                % (eos_name[cost_array.index(min(cost_array))])
            )

        elif self.eostype == "enthalpy":

            print("\nDetermining best model for each phase...")

            self.selected_coeffs = []
            self.model = []

            for i, ifile in enumerate(self.infiles):

                # Vinet
                def residuals_vinet(coeffs, y, x):
                    return y - self.eos_vinet(coeffs, x)

                self.eos_vinet_fitted = least_squares(
                    residuals_vinet,
                    self.coeffs0[i],
                    args=(self.energy[i], self.volume[i]),
                )

                # Birch
                def residuals_birch(coeffs, y, x):
                    return y - self.eos_birch(coeffs, x)

                self.eos_birch_fitted = least_squares(
                    residuals_birch,
                    self.coeffs0[i],
                    args=(self.energy[i], self.volume[i]),
                )

                # Murnaghan
                def residuals_murnaghan(coeffs, y, x):
                    return y - self.eos_murnaghan(coeffs, x)

                self.eos_murnaghan_fitted = least_squares(
                    residuals_murnaghan,
                    self.coeffs0[i],
                    args=(self.energy[i], self.volume[i]),
                )

                # Birch-Murnaghan
                def residuals_birch_murnaghan(coeffs, y, x):
                    return y - self.eos_birch_murnaghan(coeffs, x)

                self.eos_birch_murnaghan_fitted = least_squares(
                    residuals_birch_murnaghan,
                    self.coeffs0[i],
                    args=(self.energy[i], self.volume[i]),
                )

                eos_name = ["Vinet", "Birch", "Murnaghan", "Birch-Murnaghan"]

                cost_array = [
                    np.mean(self.eos_vinet_fitted.fun ** 2),
                    np.mean(self.eos_birch_fitted.fun ** 2),
                    np.mean(self.eos_murnaghan_fitted.fun ** 2),
                    np.mean(self.eos_birch_murnaghan_fitted.fun ** 2),
                ]

                coeffs_array = [
                    self.eos_vinet_fitted.x,
                    self.eos_birch_fitted.x,
                    self.eos_murnaghan_fitted.x,
                    self.eos_birch_murnaghan_fitted.x,
                ]

                print(
                    "%s :  %s " % (ifile, eos_name[cost_array.index(min(cost_array))])
                )

                self.selected_coeffs.append(
                    coeffs_array[cost_array.index(min(cost_array))]
                )
                self.model.append(eos_name[cost_array.index(min(cost_array))])

    ####################### EOS MODELS for Energy ##############################

    # Murnaghan
    def eos_murnaghan(self, coeffs, vol):
        "Ref: Phys. Rev. B 28, 5480 (1983)"
        E0, B0, Bp, V0 = coeffs
        E = (
            E0
            + B0 / Bp * vol * ((V0 / vol) ** Bp / (Bp - 1.0) + 1.0)
            - V0 * B0 / (Bp - 1.0)
        )
        return E

    # Birch-Murnaghan
    def eos_birch_murnaghan(self, coeffs, vol):
        "Ref: Phys. Rev. B 70, 224107"
        E0, B0, Bp, V0 = coeffs
        eta = (vol / V0) ** (1.0 / 3.0)
        E = E0 + 9.0 * B0 * V0 / 16.0 * (eta ** 2 - 1.0) ** 2 * (
            6.0 + Bp * (eta ** 2 - 1.0) - 4.0 * eta ** 2
        )
        return E

    # Birch
    def eos_birch(self, coeffs, vol):
        """
        Ref: Michael J. Mehl; Barry M. Klein; Dimitris A. Papaconstantopoulos. First-Principles Calculation of Elastic Properties. In Intermetallic Compounds; John Wiley & Sons Ltd, 1994; Vol. 1.

        """
        E0, B0, Bp, V0 = coeffs
        E = (
            E0
            + 9.0 / 8.0 * B0 * V0 * ((V0 / vol) ** (2.0 / 3.0) - 1.0) ** 2
            + 9.0 / 16.0 * B0 * V0 * (Bp - 4.0) * ((V0 / vol) ** (2.0 / 3.0) - 1.0) ** 3
        )
        return E

    # Vinet
    def eos_vinet(self, coeffs, vol):
        "Ref: Phys. Rev. B 70, 224107"
        E0, B0, Bp, V0 = coeffs
        eta = (vol / V0) ** (1.0 / 3.0)
        E = E0 + 2.0 * B0 * V0 / (Bp - 1.0) ** 2 * (
            2.0
            - (5.0 + 3.0 * Bp * (eta - 1.0) - 3.0 * eta)
            * np.exp(-3.0 * (Bp - 1.0) * (eta - 1.0) / 2.0)
        )
        return E

    ####################### EOS MODELS for Pressure ##############################

    # Birch-Murnaghan Pressure EOS
    def eos_birch_murnaghan_pressure(self, coeffs, vol):
        """
        Ref: doi:10.3390/min9120745
        """
        P0, K0, Kp, V0 = coeffs
        eta = (V0 / vol) ** (1.0 / 3.0)

        P = (
            (3 / 2)
            * K0
            * (eta ** 7 - eta ** 5)
            * (1 + 0.75 * (Kp - 4) * (eta ** 2 - 1))
        )
        return P

    # Outputting data to files

    def to_dict(self):
        if self.eostype == "enthalpy":
            return {
                "H": self.H.tolist(),
                "au": self.au,
                "coeffs0": self.coeffs0.tolist(),
                "deltaH_index": self.deltaH_index,
                "energy": self.energy.tolist(),
                "eos_birch_fitted_cost": self.eos_birch_fitted["cost"],
                "eos_birch_murnaghan_fitted_cost": self.eos_birch_murnaghan_fitted[
                    "cost"
                ],
                "eos_murnaghan_fitted_cost": self.eos_murnaghan_fitted["cost"],
                "eos_vinet_fitted_cost": self.eos_vinet_fitted["cost"],
                "eostype": self.eostype,
                "infiles": self.infiles,
                "labels": self.labels,
                "model": self.model,
                "natoms": self.natoms,
                "pressure": self.pressure.tolist(),
                "selected_coeffs": [i.tolist() for i in self.selected_coeffs],
                "trans_mat_reordered": self.trans_mat_reordered.tolist(),
                "vlim": self.vlim,
                "vlim_list": self.vlim_list,
                "vol_array": self.vol_array.tolist(),
                "volume": self.volume.tolist(),
            }
        else:
            return {
                "P_birch": self.P_birch.tolist(),
                "P_birch_murnaghan": self.P_birch_murnaghan.tolist(),
                "P_murnaghan": self.P_murnaghan.tolist(),
                "P_vinet": self.P_vinet.tolist(),
                "au": self.au,
                "coeffs0": self.coeffs0,
                "energy": self.energy.tolist(),
                "eos_birch_fitted_cost": self.eos_birch_fitted["cost"],
                "eos_birch_murnaghan_fitted_cost": self.eos_birch_murnaghan_fitted[
                    "cost"
                ],
                "eos_murnaghan_fitted_cost": self.eos_murnaghan_fitted["cost"],
                "eos_vinet_fitted_cost": self.eos_vinet_fitted["cost"],
                "eostype": self.eostype,
                "model": self.model,
                "natoms": self.natoms,
                "raw_data": self.raw_data,
                "vlim": self.vlim,
                "vol_array": self.vol_array.tolist(),
                "volume": self.volume.tolist(),
            }

    def to_json(self, outfile="eos.json"):

        wf = open(outfile, "w")
        json.dump(self.to_dict(), wf, sort_keys=True, indent=4, separators=(",", ": "))
        wf.close()

    def to_xml(self, outfile="eos.xml"):
        wf = open(outfile, "w")
        xml = dicttoxml(self.to_dict())
        dom = parseString(xml)
        wf.write(dom.toprettyxml())
        wf.close()



#!/usr/bin/env python

from ..tests import stability
import spglib
import numpy as np

crystallist = np.array(
    [
        "cubic",
        "hexagonal",
        "tetragonal",
        "rhombohedral-1",
        "rhombohedral-2",
        "orthorhombic",
        "monoclinic",
    ]
)

latticelist = np.array(
    [
        "hexagonal",
        "square",
        "rectangular",
        "rectangular-center",
        "oblique",
    ]
)


def crystal_select(cnew=None, cell=None, crystal_type=None, verbose=True):
    """This method selects crystal types."""

    to_print = ""
    to_print += "\n------------------------------------------------------------------"
    to_print += "Mechanical Stability Tests"
    to_print += "------------------------------------------------------------------\n"

    if crystal_type is not None:
        stability.stability_test(cnew, crystal_type)

    else:
        to_print += "WARNING: crystal symmetry class  was not provided by user, it will be taken from the OUTCAR"
        to_print += "One of the following was expected as the second argument: \n 'cubic', 'hexagonal', 'tetragonal', 'rhombohedral-1', 'rhombohedral-2', 'orthorhombic', 'monoclinic'"
        crystal_type = ""
        spg = int(spglib.get_spacegroup(cell, symprec=1e-5).split()[1][1:-1])
        if spg >= 1 and spg <= 2:
            crystal_type = "triclinic"
        if spg >= 3 and spg <= 15:
            crystal_type = "monoclinic"
        if spg >= 16 and spg <= 74:
            crystal_type = "orthorhombic"
        if spg >= 75 and spg <= 142:
            crystal_type = "tetragonal"
        if spg >= 143 and spg <= 167:
            if spg == 155 or spg == 160 or spg == 166:
                crystal_type = "rhombohedral-1"
            else:
                crystal_type = "rhombohedral-2"
        if spg >= 168 and spg <= 194:
            crystal_type = "hexagonal"
        if spg >= 195:
            crystal_type = "cubic"

        to_print += "From OUTCAR the crystal type is = %s" % crystal_type
        if verbose:
            print(to_print)
        stability.stability_test(cnew, crystal_type, verbose=verbose)


def lattice_select(cnew=None, cell=None, lattice_type=None, verbose=True):
    """This method selects crystal types."""
    to_print = ""
    if lattice_type is not None:

        to_print += "\n \t \t Mechanical Stability Test \n"
        stability.stability_test_2d(cnew, lattice_type)

    else:
        to_print += "\n"
        to_print += "WARNING: crystal symmetry class  was not provided by user"
        to_print += "One of the following was expected as the second argument: \n 'hexagonal', 'square', 'rectangular', 'rectangular-center', 'oblique'"
    if verbose:
        print(to_print)




# THIS FILE IS GENERATED FROM MECHELASTIC SETUP.PY.
name = 'MechElastic'
version = '1.2.16'
description = 'A Python library to calculate elastic properties of materials. '
url = 'https://github.com/romerogroup/MechElastic'
author = 'Sobhit Singh, Logan Lang, Viviana Dovale-Farelo, Uthpala Herath, Pedram Tavadze, François-Xavier Coudert, Aldo H. Romero'
email = 'smsingh@mail.wvu.edu'
status = 'development'
copyright = 'Copyright 2021'
date = 'Apr 24th, 2021'



# -*- coding: utf-8 -*-

import json
import math
import os
import platform
import random
import re
import sys
import time


from collections import OrderedDict
from io import StringIO
import requests

import matplotlib.pyplot as plt
import numpy as np
from scipy import optimize


__author__ = "Romain Gaillac and François-Xavier Coudert"
__version__ = "2019.01.09"
__license__ = "MIT"


def make3DPlot(func, legend="", npoints=100):

    str1 = legend.split("'")[0]
    str2 = legend.split("'")[1]

    u = np.linspace(0, np.pi, npoints)
    v = np.linspace(0, 2 * np.pi, 2 * npoints)
    r = np.zeros(len(u) * len(v))

    dataX = [[0.0 for i in range(len(v))] for j in range(len(u))]
    dataY = [[0.0 for i in range(len(v))] for j in range(len(u))]
    dataZ = [[0.0 for i in range(len(v))] for j in range(len(u))]
    dataR = [["0.0" for i in range(len(v))] for j in range(len(u))]

    count = 0
    for cu in range(len(u)):
        for cv in range(len(v)):

            r_tmp = func(u[cu], v[cv])
            z = r_tmp * np.cos(u[cu])
            x = r_tmp * np.sin(u[cu]) * np.cos(v[cv])
            y = r_tmp * np.sin(u[cu]) * np.sin(v[cv])
            dataX[cu][cv] = x
            dataY[cu][cv] = y
            dataZ[cu][cv] = z
            dataR[cu][cv] = (
                "'E = "
                + str(float(int(10 * r_tmp)) / 10.0)
                + " GPa, "
                + "\u03B8 = "
                + str(float(int(10 * u[cu] * 180 / np.pi)) / 10.0)
                + "\u00B0, "
                + "\u03c6 = "
                + str(float(int(10 * v[cv] * 180 / np.pi)) / 10.0)
                + "\u00B0'"
            )
            count = count + 1
    return (dataX, dataY, dataZ, dataR)


def make3DPlotPosNeg(func, legend="", npoints=100):

    u = np.linspace(0, np.pi, npoints)
    v = np.linspace(0, 2 * np.pi, 2 * npoints)

    dataX1 = [[0.0 for i in range(len(v))] for j in range(len(u))]
    dataY1 = [[0.0 for i in range(len(v))] for j in range(len(u))]
    dataZ1 = [[0.0 for i in range(len(v))] for j in range(len(u))]
    dataR1 = [["0.0" for i in range(len(v))] for j in range(len(u))]

    count = 0
    for cu in range(len(u)):
        for cv in range(len(v)):
            r_tmp = max(0, func(u[cu], v[cv]))
            z = r_tmp * np.cos(u[cu])
            x = r_tmp * np.sin(u[cu]) * np.cos(v[cv])
            y = r_tmp * np.sin(u[cu]) * np.sin(v[cv])
            dataX1[cu][cv] = x
            dataY1[cu][cv] = y
            dataZ1[cu][cv] = z
            dataR1[cu][cv] = (
                "'"
                + "\u03B2 = "
                + str(float(int(10 * r_tmp)) / 10.0)
                + " TPa'"
                + "+'-1'.sup()+"
                + "', \u03B8 = "
                + str(float(int(10 * u[cu] * 180 / np.pi)) / 10.0)
                + "\u00B0, "
                + "\u03c6 = "
                + str(float(int(10 * v[cv] * 180 / np.pi)) / 10.0)
                + "\u00B0'"
            )
            count = count + 1

    dataX2 = [[0.0 for i in range(len(v))] for j in range(len(u))]
    dataY2 = [[0.0 for i in range(len(v))] for j in range(len(u))]
    dataZ2 = [[0.0 for i in range(len(v))] for j in range(len(u))]
    dataR2 = [["0.0" for i in range(len(v))] for j in range(len(u))]

    count = 0
    for cu in range(len(u)):
        for cv in range(len(v)):
            r_tmp = max(0, -func(u[cu], v[cv]))
            z = r_tmp * np.cos(u[cu])
            x = r_tmp * np.sin(u[cu]) * np.cos(v[cv])
            y = r_tmp * np.sin(u[cu]) * np.sin(v[cv])
            dataX2[cu][cv] = x
            dataY2[cu][cv] = y
            dataZ2[cu][cv] = z
            dataR2[cu][cv] = (
                "'"
                + "\u03B2 = -"
                + str(float(int(10 * r_tmp)) / 10.0)
                + " TPa'"
                + "+'-1'.sup()+"
                + "', \u03B8 = "
                + str(float(int(10 * u[cu] * 180 / np.pi)) / 10.0)
                + "\u00B0, "
                + "\u03c6 = "
                + str(float(int(10 * v[cv] * 180 / np.pi)) / 10.0)
                + "\u00B0'"
            )
            count = count + 1

    return ((dataX1, dataY1, dataZ1, dataR1), (dataX2, dataY2, dataZ2, dataR2))


def make3DPlot2(func, legend="", npoints=200):

    u = np.linspace(0, np.pi, npoints)
    v = np.linspace(0, np.pi, npoints)
    w = [v[i] + np.pi for i in range(1, len(v))]
    v = np.append(v, w)

    dataX1 = [[0.0 for i in range(len(v))] for j in range(len(u))]
    dataY1 = [[0.0 for i in range(len(v))] for j in range(len(u))]
    dataZ1 = [[0.0 for i in range(len(v))] for j in range(len(u))]
    dataR1 = [["0.0" for i in range(len(v))] for j in range(len(u))]

    dataX2 = [[0.0 for i in range(len(v))] for j in range(len(u))]
    dataY2 = [[0.0 for i in range(len(v))] for j in range(len(u))]
    dataZ2 = [[0.0 for i in range(len(v))] for j in range(len(u))]
    dataR2 = [["0.0" for i in range(len(v))] for j in range(len(u))]

    count = 0
    r = [0.0, 0.0, np.pi / 2.0, np.pi / 2.0]
    for cu in range(len(u)):
        for cv in range(len(v)):

            r = func(u[cu], v[cv], r[2], r[3])
            z = np.cos(u[cu])
            x = np.sin(u[cu]) * np.cos(v[cv])
            y = np.sin(u[cu]) * np.sin(v[cv])

            r1_tmp = r[0]
            z1 = r1_tmp * z
            x1 = r1_tmp * x
            y1 = r1_tmp * y
            dataX1[cu][cv] = x1
            dataY1[cu][cv] = y1
            dataZ1[cu][cv] = z1
            dataR1[cu][cv] = (
                "'"
                + "G'"
                + "+'min'.sub()+"
                + "' = "
                + str(float(int(10 * r1_tmp)) / 10.0)
                + "GPa, "
                + "\u03B8 = "
                + str(float(int(10 * u[cu] * 180 / np.pi)) / 10.0)
                + "\u00B0, "
                + "\u03c6 = "
                + str(float(int(10 * v[cv] * 180 / np.pi)) / 10.0)
                + "\u00B0'"
            )

            r2_tmp = r[1]
            z2 = r2_tmp * z
            x2 = r2_tmp * x
            y2 = r2_tmp * y
            dataX2[cu][cv] = x2
            dataY2[cu][cv] = y2
            dataZ2[cu][cv] = z2
            dataR2[cu][cv] = (
                "'"
                + "G'"
                + "+'max'.sub()+"
                + "' = "
                + str(float(int(10 * r1_tmp)) / 10.0)
                + "GPa, "
                + "\u03B8 = "
                + str(float(int(10 * u[cu] * 180 / np.pi)) / 10.0)
                + "\u00B0, "
                + "\u03c6 = "
                + str(float(int(10 * v[cv] * 180 / np.pi)) / 10.0)
                + "\u00B0'"
            )
            count = count + 1

    i = random.randint(0, 100000)
    return ((dataX1, dataY1, dataZ1, dataR1), (dataX2, dataY2, dataZ2, dataR2))


def make3DPlot3(func, legend="", width=600, height=600, npoints=100):

    str1 = legend.split("'")[0]
    str2 = legend.split("'")[1]

    u = np.linspace(0, np.pi, npoints)
    v = np.linspace(0, np.pi, npoints)
    w = [v[i] + np.pi for i in range(1, len(v))]
    v = np.append(v, w)

    dataX1 = [[0.0 for i in range(len(v))] for j in range(len(u))]
    dataY1 = [[0.0 for i in range(len(v))] for j in range(len(u))]
    dataZ1 = [[0.0 for i in range(len(v))] for j in range(len(u))]
    dataR1 = [["0.0" for i in range(len(v))] for j in range(len(u))]

    dataX2 = [[0.0 for i in range(len(v))] for j in range(len(u))]
    dataY2 = [[0.0 for i in range(len(v))] for j in range(len(u))]
    dataZ2 = [[0.0 for i in range(len(v))] for j in range(len(u))]
    dataR2 = [["0.0" for i in range(len(v))] for j in range(len(u))]

    dataX3 = [[0.0 for i in range(len(v))] for j in range(len(u))]
    dataY3 = [[0.0 for i in range(len(v))] for j in range(len(u))]
    dataZ3 = [[0.0 for i in range(len(v))] for j in range(len(u))]
    dataR3 = [["0.0" for i in range(len(v))] for j in range(len(u))]

    count = 0
    r = [0.0, 0.0, 0.0, np.pi / 2.0, np.pi / 2.0]
    ruv = [[r for i in range(len(u))] for j in range(len(v))]
    for cu in range(len(u)):
        for cv in range(len(v)):
            ruv[cv][cu] = func(u[cu], v[cv], r[3], r[4])

    for cu in range(len(u)):
        for cv in range(len(v)):

            z = np.cos(u[cu])
            x = np.sin(u[cu]) * np.cos(v[cv])
            y = np.sin(u[cu]) * np.sin(v[cv])

            r = ruv[cv][cu]
            r1_tmp = r[0]
            dataX1[cu][cv] = r1_tmp * x
            dataY1[cu][cv] = r1_tmp * y
            dataZ1[cu][cv] = r1_tmp * z
            dataR1[cu][cv] = (
                "'"
                + "\u03BD'"
                + "+'min'.sub()+"
                + "' = "
                + str(float(int(100 * r1_tmp)) / 100.0)
                + ", "
                + "\u03B8 = "
                + str(float(int(100 * u[cu] * 180 / np.pi)) / 100.0)
                + "\u00B0, "
                + "\u03c6 = "
                + str(float(int(100 * v[cv] * 180 / np.pi)) / 100.0)
                + "\u00B0'"
            )

            r2_tmp = r[1]
            dataX2[cu][cv] = r2_tmp * x
            dataY2[cu][cv] = r2_tmp * y
            dataZ2[cu][cv] = r2_tmp * z
            dataR2[cu][cv] = float(int(100 * r2_tmp)) / 100.0

            r3_tmp = r[2]
            dataX3[cu][cv] = r3_tmp * x
            dataY3[cu][cv] = r3_tmp * y
            dataZ3[cu][cv] = r3_tmp * z
            dataR3[cu][cv] = (
                "'"
                + "\u03BD'"
                + "+'max'.sub()+"
                + "' = "
                + str(float(int(100 * r3_tmp)) / 100.0)
                + ", "
                + "\u03B8 = "
                + str(float(int(100 * u[cu] * 180 / np.pi)) / 100.0)
                + "\u00B0, "
                + "\u03c6 = "
                + str(float(int(100 * v[cv] * 180 / np.pi)) / 100.0)
                + "\u00B0'"
            )
            count = count + 1

    return (
        (dataX1, dataY1, dataZ1, dataR1),
        (dataX2, dataY2, dataZ2, dataR2),
        (dataX3, dataY3, dataZ3, dataR3),
    )


# Polar plot functions
################################################################################################


def makePolarPlot(func, legend="", p="xy", npoints=200):
    u = np.linspace(0, 2 * np.pi, npoints)
    r = list(map(func, u))
    if p == "xy":
        x = r * np.cos(u)
        y = r * np.sin(u)
    else:
        y = r * np.cos(u)
        x = r * np.sin(u)

    return (x, y)


def makePolarPlotPosNeg(func, legend="", p="xy", npoints=200):

    u = np.linspace(0, 2 * np.pi, npoints)
    r = list(map(lambda x: max(0, func(x)), u))
    if p == "xy":
        x1 = r * np.cos(u)
        y1 = r * np.sin(u)
    else:
        y1 = r * np.cos(u)
        x1 = r * np.sin(u)
    r = list(map(lambda x: max(0, -func(x)), u))
    if p == "xy":
        x2 = r * np.cos(u)
        y2 = r * np.sin(u)
    else:
        y2 = r * np.cos(u)
        x2 = r * np.sin(u)

    return ((x1, y1), (x2, y2))


def makePolarPlot2(func, legend="", p="xy", npoints=200):

    u = np.linspace(0, 2 * np.pi, npoints)
    r = list(map(func, u))

    if p == "xy":
        x1 = np.array([ir[0] * np.cos(iu) for ir, iu in zip(r, u)])
        y1 = np.array([ir[0] * np.sin(iu) for ir, iu in zip(r, u)])
        x2 = np.array([ir[1] * np.cos(iu) for ir, iu in zip(r, u)])
        y2 = np.array([ir[1] * np.sin(iu) for ir, iu in zip(r, u)])
    else:
        y1 = np.array([ir[0] * np.cos(iu) for ir, iu in zip(r, u)])
        x1 = np.array([ir[0] * np.sin(iu) for ir, iu in zip(r, u)])
        y2 = np.array([ir[1] * np.cos(iu) for ir, iu in zip(r, u)])
        x2 = np.array([ir[1] * np.sin(iu) for ir, iu in zip(r, u)])

    return ((x1, y1), (x2, y2))


def makePolarPlot3(func, legend="", p="xy", npoints=200):

    u = np.linspace(0, 2 * np.pi, npoints)
    r = list(map(func, u))

    if p == "xy":
        x1 = np.array([ir[0] * np.cos(iu) for ir, iu in zip(r, u)])
        y1 = np.array([ir[0] * np.sin(iu) for ir, iu in zip(r, u)])
        x2 = np.array([ir[1] * np.cos(iu) for ir, iu in zip(r, u)])
        y2 = np.array([ir[1] * np.sin(iu) for ir, iu in zip(r, u)])
        x3 = np.array([ir[2] * np.cos(iu) for ir, iu in zip(r, u)])
        y3 = np.array([ir[2] * np.sin(iu) for ir, iu in zip(r, u)])
    else:
        y1 = np.array([ir[0] * np.cos(iu) for ir, iu in zip(r, u)])
        x1 = np.array([ir[0] * np.sin(iu) for ir, iu in zip(r, u)])
        y2 = np.array([ir[1] * np.cos(iu) for ir, iu in zip(r, u)])
        x2 = np.array([ir[1] * np.sin(iu) for ir, iu in zip(r, u)])
        y3 = np.array([ir[2] * np.cos(iu) for ir, iu in zip(r, u)])
        x3 = np.array([ir[2] * np.sin(iu) for ir, iu in zip(r, u)])

    return ((x1, y1), (x2, y2), (x3, y3))


########################################################################################################


def dirVec(theta, phi):
    return [
        math.sin(theta) * math.cos(phi),
        math.sin(theta) * math.sin(phi),
        math.cos(theta),
    ]


def dirVec1(theta, phi, chi):
    return [
        math.sin(theta) * math.cos(phi),
        math.sin(theta) * math.sin(phi),
        math.cos(theta),
    ]


def dirVec2(theta, phi, chi):
    return [
        math.cos(theta) * math.cos(phi) * math.cos(chi) - math.sin(phi) * math.sin(chi),
        math.cos(theta) * math.sin(phi) * math.cos(chi) + math.cos(phi) * math.sin(chi),
        -math.sin(theta) * math.cos(chi),
    ]


# Functions to minimize/maximize
def minimize(func, dim):
    if dim == 2:
        r = ((0, np.pi), (0, np.pi))
        n = 25
    elif dim == 3:
        r = ((0, np.pi), (0, np.pi), (0, np.pi))
        n = 10

    # TODO -- try basin hopping or annealing
    return optimize.brute(func, r, Ns=n, full_output=True, finish=optimize.fmin)[0:2]


def maximize(func, dim):
    res = minimize(lambda x: -func(x), dim)
    return (res[0], -res[1])


class ELATE:
    def __init__(self, s,density = None):

        self.elas = Elastic(s)
        self.elasList = s
        self.density = density
        minE = minimize(self.elas.Young, 2)
        maxE = maximize(self.elas.Young, 2)
        minLC = minimize(self.elas.LC, 2)
        maxLC = maximize(self.elas.LC, 2)
        minG = minimize(self.elas.shear, 3)
        maxG = maximize(self.elas.shear, 3)
        minNu = minimize(self.elas.Poisson, 3)
        maxNu = maximize(self.elas.Poisson, 3)

        self.voigtE = self.elas.averages()[0][0]
        self.reussE = self.elas.averages()[1][0]
        self.hillE = self.elas.averages()[2][0]
        self.max_E = maxE[1]
        self.min_E = minE[1]
        self.min_axis_E = tuple(dirVec(*minE[0]))
        self.max_axis_E = tuple(dirVec(*maxE[0]))
        self.anis_E = maxE[1] / minE[1]

        self.voigtLC = self.elas.averages()[0][1]
        self.reussLC = self.elas.averages()[1][1]
        self.hillLC = self.elas.averages()[2][1]
        self.max_LC = maxLC[1]
        self.min_LC = minLC[1]
        self.min_axis_LC = tuple(dirVec(*minLC[0]))
        self.max_axis_LC = tuple(dirVec(*maxLC[0]))
        if minLC[1] > 0:
            self.anis_LC = maxLC[1] / minLC[1]
        else:
            self.anis_LC = "&infin;"

        self.voigtShear = self.elas.averages()[0][2]
        self.reussShear = self.elas.averages()[1][2]
        self.hillShear = self.elas.averages()[2][2]
        self.max_Shear = maxG[1]
        self.min_Shear = minG[1]
        self.min_axis_Shear = tuple(dirVec1(*minG[0]))
        self.max_axis_Shear = tuple(dirVec1(*maxG[0]))
        self.mix_2nd_axis_Shear = tuple(dirVec2(*minG[0]))
        self.max_2nd_axis_Shear = tuple(dirVec2(*maxG[0]))
        self.anis_Shear = maxG[1] / minG[1]

        self.voigtPoisson = self.elas.averages()[0][3]
        self.reussPoisson = self.elas.averages()[1][3]
        self.hillPoisson = self.elas.averages()[2][3]
        self.max_Poisson = maxNu[1]
        self.min_Poisson = minNu[1]
        self.min_axis_Poisson = tuple(dirVec1(*minNu[0]))
        self.max_axis_Poisson = tuple(dirVec1(*maxNu[0]))
        self.min_2nd_axis_Poisson = tuple(dirVec2(*minNu[0]))
        self.max_2nd_axis_Poisson = tuple(dirVec2(*maxNu[0]))
        if minNu[1] * maxNu[1] > 0:
            self.anis_Poisson = maxNu[1] / minNu[1]
        else:
            self.anix_Poisson = "&infin;"

    def YOUNG2D(self, npoints):
        data1 = makePolarPlot(
            lambda x: self.elas.Young([np.pi / 2, x]),
            "Young's modulus in (xy) plane",
            "xy",
            npoints=npoints,
        )
        data2 = makePolarPlot(
            lambda x: self.elas.Young([x, 0]),
            "Young's modulus in (xz) plane",
            "xz",
            npoints=npoints,
        )
        data3 = makePolarPlot(
            lambda x: self.elas.Young([x, np.pi / 2]),
            "Young's modulus in (yz) plane",
            "yz",
            npoints=npoints,
        )
        return (data1, data2, data3)

    def YOUNG3D(self, npoints):

        if self.elas.isOrthorhombic():
            self.elas = ElasticOrtho(self.elas)

        data = make3DPlot(
            lambda x, y: self.elas.Young_2(x, y), "Young's modulus", npoints=npoints
        )

        return data

    def LC2D(self, npoints):
        data1 = makePolarPlotPosNeg(
            lambda x: self.elas.LC([np.pi / 2, x]),
            "linear compressibility in (xy) plane",
            "xy",
            npoints=npoints,
        )
        data2 = makePolarPlotPosNeg(
            lambda x: self.elas.LC([x, 0]),
            "linear compressibility in (xz) plane",
            "xz",
            npoints=npoints,
        )
        data3 = makePolarPlotPosNeg(
            lambda x: self.elas.LC([x, np.pi / 2]),
            "linear compressibility in (yz) plane",
            "yz",
            npoints=npoints,
        )
        return (data1, data2, data3)

    def LC3D(self, npoints):

        if self.elas.isOrthorhombic():
            self.elas = ElasticOrtho(self.elas)

        data = make3DPlotPosNeg(
            lambda x, y: self.elas.LC_2(x, y), "Linear compressiblity", npoints=npoints
        )

        return data

    def SHEAR2D(self, npoints):
        data1 = makePolarPlot2(
            lambda x: self.elas.shear2D([np.pi / 2, x]),
            "Shear modulus in (xy) plane",
            "xy",
            npoints=npoints,
        )
        data2 = makePolarPlot2(
            lambda x: self.elas.shear2D([x, 0]), "Shear modulus in (xz) plane", "xz"
        )
        data3 = makePolarPlot2(
            lambda x: self.elas.shear2D([x, np.pi / 2]),
            "Shear modulus in (yz) plane",
            "yz",
            npoints=npoints,
        )
        return (data1, data2, data3)

    def SHEAR3D(self, npoints):

        if self.elas.isOrthorhombic():
            self.elas = ElasticOrtho(self.elas)

        data = make3DPlot2(
            lambda x, y, g1, g2: self.elas.shear3D(x, y, g1, g2),
            "Shear modulus",
            npoints=npoints,
        )

        return data

    def POISSON2D(self, npoints):
        data1 = makePolarPlot3(
            lambda x: self.elas.Poisson2D([np.pi / 2, x]),
            "Poisson's ratio in (xy) plane",
            "xy",
            npoints=npoints,
        )
        data2 = makePolarPlot3(
            lambda x: self.elas.Poisson2D([x, 0]),
            "Poisson's ratio in (xz) plane",
            "xz",
            npoints=npoints,
        )
        data3 = makePolarPlot3(
            lambda x: self.elas.Poisson2D([x, np.pi / 2]),
            "Poisson's ratio in (yz) plane",
            "yz",
            npoints=npoints,
        )
        return (data1, data2, data3)

    def POISSON3D(self, npoints):

        if self.elas.isOrthorhombic():
            self.elas = ElasticOrtho(self.elas)

        data = make3DPlot3(
            lambda x, y, g1, g2: self.elas.poisson3D(x, y, g1, g2),
            "Poisson's ratio",
            npoints=npoints,
        )

        return data

    def COMPRESSION_SPEED2D(self, npoints):
        data1 = makePolarPlot2(
            lambda x: self.elas.compressionSpeed2D([np.pi / 2, x],density = self.density),
            "Compression Speed in (xy) plane",
            "xy",
            npoints=npoints,
        )
        data2 = makePolarPlot2(
            lambda x: self.elas.compressionSpeed2D([x, 0],density = self.density), "Compression Speed in (xz) plane", "xz"
        )
        data3 = makePolarPlot2(
            lambda x: self.elas.compressionSpeed2D([x, np.pi / 2], density = self.density),
            "Compression Speed in (yz) plane",
            "yz",
            npoints=npoints,
        )
        return (data1, data2, data3)

    def COMPRESSION_SPEED3D(self, npoints):

        if self.elas.isOrthorhombic():
            self.elas = ElasticOrtho(self.elas)

        data = make3DPlot2(
            lambda x, y, g1, g2: self.elas.compressionSpeed3D(x, y, g1, g2, density = self.density),
            "Compression speed",
            npoints=npoints,
        )

        return data

    def SHEAR_SPEED2D(self, npoints):
        data1 = makePolarPlot2(
            lambda x: self.elas.shearSpeed2D([np.pi / 2, x],density = self.density),
            "Shear speed in (xy) plane",
            "xy",
            npoints=npoints,
        )
        data2 = makePolarPlot2(
            lambda x: self.elas.shearSpeed2D([x, 0],density = self.density), "Shear speed in (xz) plane", "xz"
        )
        data3 = makePolarPlot2(
            lambda x: self.elas.shearSpeed2D([x, np.pi / 2],density = self.density),
            "Shear speed in (yz) plane",
            "yz",
            npoints=npoints,
        )
        return (data1, data2, data3)

    def SHEAR_SPEED3D(self, npoints):

        if self.elas.isOrthorhombic():
            self.elas = ElasticOrtho(self.elas)

        data = make3DPlot2(
            lambda x, y, g1, g2: self.elas.shearSpeed3D(x, y, g1, g2, density = self.density),
            "Shear speed",
            npoints=npoints,
        )
        return data

    def RATIO_COMPRESSIONAL_SHEAR2D(self, npoints):
        data1 = makePolarPlot2(
            lambda x: self.elas. ratio_compressional_shear2D([np.pi / 2, x],density = self.density),
            "Ratio compressional/shear in (xy) plane",
            "xy",
            npoints=npoints,
        )
        data2 = makePolarPlot2(
            lambda x: self.elas. ratio_compressional_shear2D([x, 0],density = self.density), "Ratio compressional/shear in (xz) plane", "xz"
        )
        data3 = makePolarPlot2(
            lambda x: self.elas. ratio_compressional_shear2D([x, np.pi / 2],density = self.density),
            "Ratio compressional/shear in (yz) plane",
            "yz",
            npoints=npoints,
        )
        return (data1, data2, data3)

    def RATIO_COMPRESSIONAL_SHEAR3D(self, npoints):

        if self.elas.isOrthorhombic():
            self.elas = ElasticOrtho(self.elas)

        data = make3DPlot2(
            lambda x, y, g1, g2: self.elas.ratio_compressional_shear3D(x, y, g1, g2,density = self.density),
            "Ratio of compression/shear",
            npoints=npoints,
        )
        return data

    def DEBYE_SPEED2D(self, npoints):
        data1 = makePolarPlot2(
            lambda x: self.elas.debyeSpeed2D([np.pi / 2, x],density = self.density),
            "Debye speed in (xy) plane",
            "xy",
            npoints=npoints,
        )
        data2 = makePolarPlot2(
            lambda x: self.elas.debyeSpeed2D([x, 0], density = self.density), "Debye speed in (xz) plane", "xz"
        )
        data3 = makePolarPlot2(
            lambda x: self.elas.debyeSpeed2D([x, np.pi / 2],density = self.density),
            "Debye speed in (yz) plane",
            "yz",
            npoints=npoints,
        )
        return (data1, data2, data3)

    def DEBYE_SPEED3D(self, npoints):

        if self.elas.isOrthorhombic():
            self.elas = ElasticOrtho(self.elas)

        data = make3DPlot2(
            lambda x, y, g1, g2: self.elas.debyeSpeed3D(x, y, g1, g2,density = self.density),
            "Debye Speed",
            npoints=npoints,
        )

        return data

    ##############################################################################
    # Plotting functions
    #############################################################################
    def plot_3D(self, elastic_calc="", npoints=100,show = True):
        import pyvista as pv


        plotter = pv.Plotter()

        plotter.set_background(color="white")
        x = None
        y = None
        z = None
        r = None
        meshes = []
        if elastic_calc == "POISSON":
            func = self.POISSON3D(npoints=100)
            colors = ["red", "green", "blue"]
            for ix, icolor in zip(range(len(func)), colors):
                x = np.array(func[ix][0])
                y = np.array(func[ix][1])
                z = np.array(func[ix][2])
                r = np.array(func[ix][3])
                if np.all((func[ix][0] == 0)):
                    continue
                else:
                    grid = pv.StructuredGrid(x, y, z)
                    meshes.append(grid)
                    if ix == 2:
                        plotter.add_mesh(grid, opacity=0.25, color=icolor)
                    else:
                        plotter.add_mesh(grid, opacity=0.50, color=icolor)

        elif elastic_calc == "SHEAR":
            func = self.SHEAR3D(npoints=100)
            colors = ["green", "blue"]
            for ix, icolor in zip(range(len(func)), colors):
                x = np.array(func[ix][0])
                y = np.array(func[ix][1])
                z = np.array(func[ix][2])
                r = np.array(func[ix][3])
                if np.all((func[ix][0] == 0)):
                    continue
                else:
                    grid = pv.StructuredGrid(x, y, z)
                    meshes.append(grid)
                    if ix == 2:
                        plotter.add_mesh(grid, opacity=0.25, color=icolor)
                    else:
                        plotter.add_mesh(grid, opacity=0.50, color=icolor)

        elif elastic_calc == "LC":
            func = self.LC3D(npoints=100)
            colors = ["green", "red"]
            for ix, icolor in zip(range(len(func)), colors):
                x = np.array(func[ix][0])
                y = np.array(func[ix][1])
                z = np.array(func[ix][2])
                r = np.array(func[ix][3])
                if np.all((func[ix][0] == 0)):
                    continue
                else:
                    grid = pv.StructuredGrid(x, y, z)
                    meshes.append(grid)
                    if ix == 2:
                        plotter.add_mesh(grid, opacity=0.25, color=icolor)
                    else:
                        plotter.add_mesh(grid, opacity=0.50, color=icolor)

        elif elastic_calc == "YOUNG":
            func = self.YOUNG3D(npoints=100)
            colors = ["green"]
            for ix, icolor in zip(range(len(func)), colors):

                x = np.array(func[0])
                y = np.array(func[1])
                z = np.array(func[2])
                r = np.array(func[3])
                if np.all((func[0] == 0)):
                    continue
                else:
                    grid = pv.StructuredGrid(x, y, z)
                    meshes.append(grid)
                    if ix == 2:
                        plotter.add_mesh(grid, opacity=0.25, color=icolor)
                    else:
                        plotter.add_mesh(grid, opacity=0.50, color=icolor)

        if self.density != None:
            if elastic_calc == "COMPRESSION_SPEED":
                func = self.COMPRESSION_SPEED3D(npoints=100)
                colors = ["green", "blue"]
                for ix, icolor in zip(range(len(func)), colors):
                    x = np.array(func[ix][0])
                    y = np.array(func[ix][1])
                    z = np.array(func[ix][2])
                    r = np.array(func[ix][3])
                    if np.all((func[ix][0] == 0)):
                        continue
                    else:
                        grid = pv.StructuredGrid(x, y, z)
                        meshes.append(grid)
                        if ix == 2:
                            plotter.add_mesh(grid, opacity=0.25, color=icolor)
                        else:
                            plotter.add_mesh(grid, opacity=0.50, color=icolor)

            elif elastic_calc == "SHEAR_SPEED":
                func = self.SHEAR_SPEED3D(npoints=100)
                colors = ["green", "blue"]
                for ix, icolor in zip(range(len(func)), colors):
                    x = np.array(func[ix][0])
                    y = np.array(func[ix][1])
                    z = np.array(func[ix][2])
                    r = np.array(func[ix][3])
                    if np.all((func[ix][0] == 0)):
                        continue
                    else:
                        grid = pv.StructuredGrid(x, y, z)
                        meshes.append(grid)
                        if ix == 2:
                            plotter.add_mesh(grid, opacity=0.25, color=icolor)
                        else:
                            plotter.add_mesh(grid, opacity=0.50, color=icolor)

            elif elastic_calc == "RATIO_COMPRESSIONAL_SHEAR":
                func = self.RATIO_COMPRESSIONAL_SHEAR3D(npoints=100)
                colors = ["green", "blue"]
                for ix, icolor in zip(range(len(func)), colors):
                    x = np.array(func[ix][0])
                    y = np.array(func[ix][1])
                    z = np.array(func[ix][2])
                    r = np.array(func[ix][3])
                    if np.all((func[ix][0] == 0)):
                        continue
                    else:
                        grid = pv.StructuredGrid(x, y, z)
                        meshes.append(grid)
                        if ix == 2:
                            plotter.add_mesh(grid, opacity=0.25, color=icolor)
                        else:
                            plotter.add_mesh(grid, opacity=0.50, color=icolor)

            elif elastic_calc == "DEBYE_SPEED":
                func = self.DEBYE_SPEED3D(npoints=100)
                colors = ["green", "blue"]
                for ix, icolor in zip(range(len(func)), colors):
                    x = np.array(func[ix][0])
                    y = np.array(func[ix][1])
                    z = np.array(func[ix][2])
                    r = np.array(func[ix][3])
                    if np.all((func[ix][0] == 0)):
                        continue
                    else:
                        grid = pv.StructuredGrid(x, y, z)
                        meshes.append(grid)
                        if ix == 2:
                            plotter.add_mesh(grid, opacity=0.25, color=icolor)
                        else:
                            plotter.add_mesh(grid, opacity=0.50, color=icolor)

        if elastic_calc in ['DEBYE_SPEED', 'SHEAR_SPEED', 'COMPRESSIONAL_SPEED', 'RATIO_COMPRESSIONAL_SHEAR'] and self.density == None:
            print("You must specify density in kg/m^3 to produce DEBYE_SPEED, SHEAR_SPEED, COMPRESSIONAL_SPEED, and RATIO_COMPRESSIONAL_SHEAR")

        plotter.add_axes()
        if show:
            plotter.show()
        return meshes
        #plotter.show_grid(color = "black")

    def plot_2D(self, elastic_calc="all", npoints=100, apply_to_plot=None, density= None, show= True, ):
        """


        Parameters
        ----------
        elastic_calc : str, optional
            The property to be plotted. The default is "all".
        npoints : TYPE, optional
            DESCRIPTION. The default is 100.
        apply_to_plot : function, optional
            A python function to be applied to each plot. The default is None.
            The following example will create axis that go through origin
            e.g. ::

                def detail(fig, ax):
                fig.set_size_inches((11, 15), forward=True)
                ax.spines["left"].set_position("center")
                ax.spines["bottom"].set_position("center")
                ax.spines["right"].set_color("none")
                ax.spines["top"].set_color("none")
                xlim = ax.get_xlim()
                ylim = ax.get_ylim()

                rec = patches.Rectangle(
                    (xlim[0], ylim[0]),
                    (xlim[1] - xlim[0]),
                    (ylim[1] - ylim[0]),
                    linewidth=2,
                    edgecolor="grey",
                    facecolor="none",
                )

                ax.add_patch(rec)

                fig.suptitle("")
                title = ax.get_title()
                ax.get_yaxis().set_visible(True)
                ax.set_title(title, y=-0.1)
                ax.tick_params(
                    which="major", axis="y", direction="inout", width=1, length=5, rotation=0
                )

                ax.tick_params(
                    which="major", axis="x", direction="inout", width=1, length=5, rotation=0
                )

        Returns
        -------
        None.

        """

        fig = plt.figure()
        subTitles = ["XY", "XZ", "YZ"]

        if elastic_calc == "POISSON":
            func = self.POISSON2D(npoints=npoints)
            colors = ["red", "green", "blue"]
            labels = ["Poisson - Neg", "Poisson - Pos", "Poisson - Max"]
            fig.suptitle("Poisson's ratio")
            for iplane, title in zip(range(len(func)), subTitles):

                ax = fig.add_subplot(1, 3, iplane + 1)
                ax.set_title(title)
                ax.get_yaxis().set_visible(False)
                for iplot, color in zip(range(len(func[0])), colors):
                    plt.plot(
                        func[iplane][iplot][0], func[iplane][iplot][1], color=color
                    )

        elif elastic_calc == "SHEAR":
            func = self.SHEAR2D(npoints=npoints)
            colors = ["green", "blue"]
            labels = ["Shear - Max", "Shear -"]
            fig.suptitle("Shear modulus")
            for iplane, title in zip(range(len(func)), subTitles):

                ax = fig.add_subplot(1, 3, iplane + 1)
                ax.set_title(title)
                ax.get_yaxis().set_visible(False)
                for iplot, color in zip(range(len(func[0])), colors):
                    plt.plot(
                        func[iplane][iplot][0], func[iplane][iplot][1], color=color
                    )

        elif elastic_calc == "LC":
            func = self.LC2D(npoints=npoints)
            colors = ["green", "red"]
            labels = ["LC-positive", "LC-negative"]
            fig.suptitle("Linear Compression")
            for iplane, title in zip(range(len(func)), subTitles):

                ax = fig.add_subplot(1, 3, iplane + 1)
                ax.set_title(title)
                ax.get_yaxis().set_visible(False)
                for iplot, color, label in zip(range(len(func[0])), colors, labels):
                    plt.plot(
                        func[iplane][iplot][0],
                        func[iplane][iplot][1],
                        color=color,
                        label=label,
                    )

        elif elastic_calc == "YOUNG":
            func = self.YOUNG2D(npoints=npoints)
            color = "green"
            label = "YOUNG"
            fig.suptitle("Young's Modulus")
            for iplane, title in zip(range(len(func)), subTitles):
                ax = fig.add_subplot(1, 3, iplane + 1)
                ax.set_title(title)
                ax.get_yaxis().set_visible(False)
                plt.plot(func[iplane][0], func[iplane][1], color=color, label=label)

        elif elastic_calc == "all":
            func = self.YOUNG2D(npoints=npoints)
            color = "green"
            label = "YOUNG"
            # fig.suptitle("Young's Modulus")
            subTitles = [
                "Young's Modulus in (XY) plane",
                "Young's Modulus in (XZ) plane",
                "Young's Modulus in (YZ) plane",
            ]
            for iplane, title in zip(range(len(func)), subTitles):
                ax = fig.add_subplot(4, 3, iplane + 1)
                ax.set_title(title)
                ax.get_yaxis().set_visible(False)
                plt.plot(func[iplane][0], func[iplane][1], color=color, label=label)
                if apply_to_plot is not None:
                    apply_to_plot(fig, ax)
            func = self.LC2D(npoints=npoints)
            colors = ["green", "red"]
            labels = ["LC-positive", "LC-negative"]
            # fig.suptitle("Linear Compression")
            subTitles = [
                "Bulk Modulus in (XY) plane",
                "Bulk Modulus in (XZ) plane",
                "Bulk Modulus in (YZ) plane",
            ]
            for iplane, title in zip(range(len(func)), subTitles):

                ax = fig.add_subplot(4, 3, iplane + 4)
                ax.set_title(title)
                ax.get_yaxis().set_visible(False)
                for iplot, color, label in zip(range(len(func[0])), colors, labels):
                    plt.plot(
                        func[iplane][iplot][0],
                        func[iplane][iplot][1],
                        color=color,
                        label=label,
                    )
                if apply_to_plot is not None:
                    apply_to_plot(fig, ax)
            func = self.SHEAR2D(npoints=npoints)
            colors = ["green", "blue"]
            labels = ["Shear - Max", "Shear -"]
            fig.suptitle("Shear modulus")
            subTitles = [
                "Shear Modulus in (XY) plane",
                "Shear Modulus in (XZ) plane",
                "Shear Modulus in (YZ) plane",
            ]
            for iplane, title in zip(range(len(func)), subTitles):

                ax = fig.add_subplot(4, 3, iplane + 7)
                ax.set_title(title)
                ax.get_yaxis().set_visible(False)
                for iplot, color in zip(range(len(func[0])), colors):
                    plt.plot(
                        func[iplane][iplot][0], func[iplane][iplot][1], color=color
                    )
                if apply_to_plot is not None:
                    apply_to_plot(fig, ax)
            func = self.POISSON2D(npoints=npoints)
            colors = ["red", "green", "blue"]
            labels = ["Poisson - Neg", "Poisson - Pos", "Poisson - Max"]
            fig.suptitle("Poisson's ratio")
            subTitles = [
                "Poisson's Ratio in (XY) plane",
                "Poisson's Ratio in (XZ) plane",
                "Poisson's Ratio in (YZ) plane",
            ]
            for iplane, title in zip(range(len(func)), subTitles):

                ax = fig.add_subplot(4, 3, iplane + 10)
                ax.set_title(title)
                ax.get_yaxis().set_visible(False)
                for iplot, color in zip(range(len(func[0])), colors):
                    plt.plot(
                        func[iplane][iplot][0], func[iplane][iplot][1], color=color
                    )
                if apply_to_plot is not None:
                    apply_to_plot(fig, ax)

        if density != None :

            if elastic_calc == "COMPRESSION_SPEED":
                func = self.COMPRESSION_SPEED2D(npoints=npoints)
                colors = ["green", "blue"]
                labels = ["Compression speed - Max", "Compression speed -"]
                fig.suptitle("Compression speed")

                for iplane, title in zip(range(len(func)), subTitles):

                    ax = fig.add_subplot(1, 3, iplane + 1)
                    ax.set_title(title)
                    ax.get_yaxis().set_visible(False)
                    for iplot, color in zip(range(len(func[0])), colors):
                        plt.plot(
                            func[iplane][iplot][0], func[iplane][iplot][1], color=color
                        )
            if elastic_calc == "SHEAR_SPEED":
                func = self.SHEAR_SPEED2D(npoints=npoints)
                colors = ["green", "blue"]
                labels = ["Shear speed - Max", "Shear speed -"]
                fig.suptitle("Shear speed")
                for iplane, title in zip(range(len(func)), subTitles):

                    ax = fig.add_subplot(1, 3, iplane + 1)
                    ax.set_title(title)
                    ax.get_yaxis().set_visible(False)
                    for iplot, color in zip(range(len(func[0])), colors):
                        plt.plot(
                            func[iplane][iplot][0], func[iplane][iplot][1], color=color
                        )
            if elastic_calc == "RATIO_COMPRESSIONAL_SHEAR":
                func = self.RATIO_COMPRESSIONAL_SHEAR2D(npoints=npoints)
                colors = ["green", "blue"]
                labels = ["Ratio compression/shear - Max", "Ratio compression/shear -"]
                fig.suptitle("Ratio compression/shear")
                for iplane, title in zip(range(len(func)), subTitles):

                    ax = fig.add_subplot(1, 3, iplane + 1)
                    ax.set_title(title)
                    ax.get_yaxis().set_visible(False)
                    for iplot, color in zip(range(len(func[0])), colors):
                        plt.plot(
                            func[iplane][iplot][0], func[iplane][iplot][1], color=color
                        )
            if elastic_calc == "DEBYE_SPEED":
                func = self.DEBYE_SPEED2D(npoints=npoints)
                colors = ["green", "blue"]
                labels = ["Debye speed - Max", "Debye speed -"]
                fig.suptitle("Debye speed")
                for iplane, title in zip(range(len(func)), subTitles):

                    ax = fig.add_subplot(1, 3, iplane + 1)
                    ax.set_title(title)
                    ax.get_yaxis().set_visible(False)
                    for iplot, color in zip(range(len(func[0])), colors):
                        plt.plot(
                            func[iplane][iplot][0], func[iplane][iplot][1], color=color
                        )
        if elastic_calc in ['DEBYE_SPEED', 'SHEAR_SPEED', 'COMPRESSIONAL_SPEED', 'RATIO_COMPRESSIONAL_SHEAR'] and density == None:
            print("You must specify density in kg/m^3 to produce DEBYE_SPEED, SHEAR_SPEED, COMPRESSIONAL_SPEED, and RATIO_COMPRESSIONAL_SHEAR")




        if show:
            plt.show()

        return fig

    def print_properties(self):
        #        print("\n \n Input: compliance matrix(coefficients in GPa) of\n")
        #        for x in self.elasList:
        #            print(str(x))

        print("\n")
        print("\n")
        print("Average properties")

        print("\n \n                         Voigt     Reuss    Hill")
        print("-------------------------------------------------------")
        print(
            "Bulk modulus   (GPa)  %9.3f %9.3f %9.3f "
            % (
                self.elas.averages()[0][0],
                self.elas.averages()[1][0],
                self.elas.averages()[2][0],
            )
        )
        print(
            "Shear modulus  (GPa)  %9.3f %9.3f %9.3f "
            % (
                self.elas.averages()[0][1],
                self.elas.averages()[1][1],
                self.elas.averages()[2][1],
            )
        )
        print(
            "Young's modulus  (GPa)  %9.3f %9.3f %9.3f "
            % (
                self.elas.averages()[0][2],
                self.elas.averages()[1][2],
                self.elas.averages()[2][2],
            )
        )
        print(
            "Poisson's ratio         %9.3f %9.3f %9.3f "
            % (
                self.elas.averages()[0][3],
                self.elas.averages()[1][3],
                self.elas.averages()[2][3],
            )
        )

        print("-------------------------------------------------------")
        # printer.printMatix(elastic_rowsList)
        print("\n")
        print("Eigenvalues of compliance matrix")
        print("\n \n   lamda_1  lamda_2  lamda_3  lamda_4  lamda_5  lamda_6")
        print("---------------------------------------------------------------")
        eigenval = sorted(np.linalg.eig(self.elas.CVoigt)[0])
        print("%9.3f %9.3f %9.3f %9.3f %9.3f %9.3f" % tuple(eigenval))
        print("--------------------------------------------------------------")
        if eigenval[0] <= 0:
            print(
                '<div class="error">Stiffness matrix is not definite positive, crystal is mechanically unstable<br/>'
            )
        print("\n")
        print("Variations of the elastic moduli")
        #        print("\n \n   Young's Modulus                Linear Compressibility                   Shear Modulus              Poisson's Ratio")
        #        print("    Emin         Emax             betaMin            betaMax               Gmin        Gmax           Nu_min      Nu_max")
        #        print("----------------------------------------------------------------------------------------------")
        minE = minimize(self.elas.Young, 2)
        maxE = maximize(self.elas.Young, 2)
        minLC = minimize(self.elas.LC, 2)
        maxLC = maximize(self.elas.LC, 2)
        minG = minimize(self.elas.shear, 3)
        maxG = maximize(self.elas.shear, 3)
        minNu = minimize(self.elas.Poisson, 3)
        maxNu = maximize(self.elas.Poisson, 3)
        # print(type((minE,maxE,minLC,maxLC,minG,maxG,minNu,maxNu)))
        # print("%9.3f GPa  %9.3f GPa  %9.3f TPa^-1   %9.3f TPa^-1    %9.3f GPa  %9.3f GPa  %9.3f  %9.3f" % (minE[1],maxE[1],minLC[1],maxLC[1],minG[1],maxG[1],minNu[1],maxNu[1]))

        anisE = "%8.3f" % (maxE[1] / minE[1])
        anisLC = ("%8.3f" % (maxLC[1] / minLC[1])) if minLC[1] > 0 else "&infin;"
        anisG = "%8.3f" % (maxG[1] / minG[1])
        anisNu = (
            ("%8.3f" % (maxNu[1] / minNu[1])) if minNu[1] * maxNu[1] > 0 else "&infin;"
        )

        # print("       %s                     %s                                %s                         %s" % (anisE,anisLC,anisG,anisNu))
        # print(dirVec1(*minG[0]))

        minEaxis = list(np.around(np.array(dirVec(*minE[0])), 3))
        maxEaxis = list(np.around(np.array(dirVec(*maxE[0])), 3))
        minLCaxis = list(np.around(np.array(dirVec(*minLC[0])), 3))
        maxLCaxis = list(np.around(np.array(dirVec(*maxLC[0])), 3))

        minGaxis = list(np.around(np.array(dirVec1(*minG[0])), 3))
        maxGaxis = list(np.around(np.array(dirVec1(*maxG[0])), 3))
        minG2ndaxis = list(np.around(np.array(dirVec2(*minG[0])), 3))
        maxG2ndaxis = list(np.around(np.array(dirVec2(*maxG[0])), 3))

        minNUaxis = list(np.around(np.array(dirVec1(*minNu[0])), 3))
        maxNUaxis = list(np.around(np.array(dirVec1(*maxNu[0])), 3))
        minNU2ndaxis = list(np.around(np.array(dirVec2(*minNu[0])), 3))
        maxNU2ndaxis = list(np.around(np.array(dirVec2(*maxNu[0])), 3))

        print("\n \n                                  Min       Max   ||   Anisotropy")
        print(
            "----------------------------------------------------------------------------------------"
        )
        print(
            "Young's Modulus     (GPa)      %9.3f %9.3f || %s "
            % (minE[1], maxE[1], anisE)
        )
        print(
            "         Min Axis:    %s      \n         Max Axis:    %s "
            % (str(tuple(minEaxis)), str(tuple(maxEaxis)))
        )
        # print(
        #     "         Min Axis:    %s      \n         Max Axis:    %s "
        #     % (str(tuple(dirVec(*minE[0]))), str(tuple(dirVec(*maxE[0]))))
        # )
        print(
            "----------------------------------------------------------------------------------------"
        )
        print(
            "Linear Compression   (TPa^-1)  %9.3f %9.3f || %s "
            % (minLC[1], maxLC[1], anisLC)
        )
        print(
            "         Min Axis:    %s      \n         Max Axis:    %s "
            % (str(tuple(minLCaxis)), str(tuple(maxLCaxis)))
        )
        print(
            "----------------------------------------------------------------------------------------"
        )
        print(
            "Shear Modulus   (GPa)          %9.3f %9.3f || %s "
            % (minG[1], maxG[1], anisG)
        )
        print(
            "         Min Axis:    %s      \n         Max Axis:    %s "
            % (str(tuple(minGaxis)), str(tuple(maxGaxis)))
        )
        print(
            "         Second Min Axis:    %s      \n         Second Max Axis:    %s "
            % (str(tuple(minG2ndaxis)), str(tuple(minG2ndaxis)))
        )
        print(
            "----------------------------------------------------------------------------------------"
        )
        print(
            "Poisson's Ratio                 %9.3f %9.3f || %s "
            % (minNu[1], maxNu[1], anisNu)
        )
        print(
            "         Min Axis:    %s      \n         Max Axis:    %s "
            % (str(tuple(minNUaxis)), str(tuple(maxNUaxis)))
        )
        print(
            "         Second Min Axis:    %s      \n         Second Max Axis:    %s "
            % (str(tuple(minNU2ndaxis)), str(tuple(maxNU2ndaxis)))
        )


#        print(
#            "Shear modulus  (GPa)  %9.3f %9.3f %9.3f "
#            % (self.elas.averages()[0][1], self.elas.averages()[1][1], self.elas.averages()[2][1])
#        )
#        print(
#            "Young modulus  (GPa)  %9.3f %9.3f %9.3f "
#            % (self.elas.averages()[0][2], self.elas.averages()[1][2], self.elas.averages()[2][2])
#        )
#        print(
#            "Poisson ratio         %9.3f %9.3f %9.3f "
#            % (self.elas.averages()[0][3], self.elas.averages()[2][3], self.elas.averages()[2][3]))
#
#        print("-------------------------------------------------------")
#        print("%9.3f %9.3f %9.3f %9.3f %9.3f %9.3f %9.3f %9.3f" % tuple(dirVec(*minE[0])) + tuple(dirVec(*maxE[0])) + tuple(dirVec(*minLC[0])) + tuple(dirVec(*maxLC[0])) + tuple(dirVec1(*minG[0])) +  tuple(dirVec1(*maxG[0])) + tuple(dirVec1(*minNu[0])) + tuple(dirVec1(*maxNu[0])))
#        print("%9.3f %9.3f %9.3f %9.3f %9.3f %9.3f %9.3f %9.3f" % tuple(dirVec2(*minG[0])) + tuple(dirVec2(*maxG[0])) + tuple(dirVec2(*minNu[0])) + tuple(dirVec2(*maxNu[0])))


class Elastic:
    """An elastic tensor, along with methods to access it"""

    def __init__(self, s ):
        """Initialize the elastic tensor from a string"""

        if not s:
            raise ValueError("no matrix was provided")

        # Argument can be a 6-line string, a list of list, or a string representation of the list of list
        try:
            if type(json.loads(s)) == list:
                s = json.loads(s)
        except:
            pass

        if type(s) == str:
            # Remove braces and pipes
            s = s.replace("|", " ").replace("(", " ").replace(")", " ")

            # Remove empty lines
            lines = [line for line in s.split("\n") if line.strip()]
            if len(lines) != 6:
                raise ValueError("should have six rows")

            # Convert to float
            try:
                mat = [list(map(float, line.split())) for line in lines]
            except:
                raise ValueError("not all entries are numbers")
        elif type(s) == list:
            # If we already have a list, simply use it
            mat = s
        else:
            raise ValueError("invalid argument as matrix")

        # Make it into a square matrix
        mat = np.array(mat)
        if mat.shape != (6, 6):
            # Is it upper triangular?
            if list(map(len, mat)) == [6, 5, 4, 3, 2, 1]:
                mat = [[0] * i + mat[i] for i in range(6)]
                mat = np.array(mat)

            # Is it lower triangular?
            if list(map(len, mat)) == [1, 2, 3, 4, 5, 6]:
                mat = [mat[i] + [0] * (5 - i) for i in range(6)]
                mat = np.array(mat)

        if mat.shape != (6, 6):
            raise ValueError("should be a square matrix")

        # Check that is is symmetric, or make it symmetric
        if np.linalg.norm(np.tril(mat, -1)) == 0:
            mat = mat + np.triu(mat, 1).transpose()
        if np.linalg.norm(np.triu(mat, 1)) == 0:
            mat = mat + np.tril(mat, -1).transpose()
        if np.linalg.norm(mat - mat.transpose()) > 1e-3:
            raise ValueError("should be symmetric, or triangular")
        elif np.linalg.norm(mat - mat.transpose()) > 0:
            mat = 0.5 * (mat + mat.transpose())

        # Store it
        self.CVoigt = mat

        # Put it in a more useful representation
        try:
            self.SVoigt = np.linalg.inv(self.CVoigt)
        except:
            raise ValueError("matrix is singular")

        VoigtMat = [[0, 5, 4], [5, 1, 3], [4, 3, 2]]

        def SVoigtCoeff(p, q):
            return 1.0 / ((1 + p // 3) * (1 + q // 3))

        self.Smat = [
            [
                [
                    [
                        SVoigtCoeff(VoigtMat[i][j], VoigtMat[k][l])
                        * self.SVoigt[VoigtMat[i][j]][VoigtMat[k][l]]
                        for i in range(3)
                    ]
                    for j in range(3)
                ]
                for k in range(3)
            ]
            for l in range(3)
        ]
        return

    def isOrthorhombic(self):
        def iszero(x):
            return abs(x) < 1.0e-3

        return (
            iszero(self.CVoigt[0][3])
            and iszero(self.CVoigt[0][4])
            and iszero(self.CVoigt[0][5])
            and iszero(self.CVoigt[1][3])
            and iszero(self.CVoigt[1][4])
            and iszero(self.CVoigt[1][5])
            and iszero(self.CVoigt[2][3])
            and iszero(self.CVoigt[2][4])
            and iszero(self.CVoigt[2][5])
            and iszero(self.CVoigt[3][4])
            and iszero(self.CVoigt[3][5])
            and iszero(self.CVoigt[4][5])
        )

    def isCubic(self):
        def iszero(x):
            return abs(x) < 1.0e-3

        return (
            iszero(self.CVoigt[0][3])
            and iszero(self.CVoigt[0][4])
            and iszero(self.CVoigt[0][5])
            and iszero(self.CVoigt[1][3])
            and iszero(self.CVoigt[1][4])
            and iszero(self.CVoigt[1][5])
            and iszero(self.CVoigt[2][3])
            and iszero(self.CVoigt[2][4])
            and iszero(self.CVoigt[2][5])
            and iszero(self.CVoigt[3][4])
            and iszero(self.CVoigt[3][5])
            and iszero(self.CVoigt[4][5])
            and iszero(self.CVoigt[0][0] - self.CVoigt[1][1])
            and iszero(self.CVoigt[0][0] - self.CVoigt[2][2])
            and iszero(self.CVoigt[0][0] - self.CVoigt[1][1])
            and iszero(self.CVoigt[0][0] - self.CVoigt[2][2])
            and iszero(self.CVoigt[3][3] - self.CVoigt[4][4])
            and iszero(self.CVoigt[3][3] - self.CVoigt[5][5])
            and iszero(self.CVoigt[0][1] - self.CVoigt[0][2])
            and iszero(self.CVoigt[0][1] - self.CVoigt[1][2])
        )

    def Young(self, x):
        a = dirVec(x[0], x[1])
        r = sum(
            [
                a[i] * a[j] * a[k] * a[l] * self.Smat[i][j][k][l]
                for i in range(3)
                for j in range(3)
                for k in range(3)
                for l in range(3)
            ]
        )
        return 1 / r

    def Young_2(self, x, y):
        a = dirVec(x, y)
        r = sum(
            [
                a[i] * a[j] * a[k] * a[l] * self.Smat[i][j][k][l]
                for i in range(3)
                for j in range(3)
                for k in range(3)
                for l in range(3)
            ]
        )
        return 1 / r

    def LC(self, x):
        a = dirVec(x[0], x[1])
        r = sum(
            [
                a[i] * a[j] * self.Smat[i][j][k][k]
                for i in range(3)
                for j in range(3)
                for k in range(3)
            ]
        )
        return 1000 * r

    def LC_2(self, x, y):
        a = dirVec(x, y)
        r = sum(
            [
                a[i] * a[j] * self.Smat[i][j][k][k]
                for i in range(3)
                for j in range(3)
                for k in range(3)
            ]
        )
        return 1000 * r

    def shear(self, x):
        a = dirVec(x[0], x[1])
        b = dirVec2(x[0], x[1], x[2])
        r = sum(
            [
                a[i] * b[j] * a[k] * b[l] * self.Smat[i][j][k][l]
                for i in range(3)
                for j in range(3)
                for k in range(3)
                for l in range(3)
            ]
        )
        return 1 / (4 * r)

    def Poisson(self, x):
        a = dirVec(x[0], x[1])
        b = dirVec2(x[0], x[1], x[2])
        r1 = sum(
            [
                a[i] * a[j] * b[k] * b[l] * self.Smat[i][j][k][l]
                for i in range(3)
                for j in range(3)
                for k in range(3)
                for l in range(3)
            ]
        )
        r2 = sum(
            [
                a[i] * a[j] * a[k] * a[l] * self.Smat[i][j][k][l]
                for i in range(3)
                for j in range(3)
                for k in range(3)
                for l in range(3)
            ]
        )
        return -r1 / r2

    def averages(self):
        A = (self.CVoigt[0][0] + self.CVoigt[1][1] + self.CVoigt[2][2]) / 3
        B = (self.CVoigt[1][2] + self.CVoigt[0][2] + self.CVoigt[0][1]) / 3
        C = (self.CVoigt[3][3] + self.CVoigt[4][4] + self.CVoigt[5][5]) / 3
        a = (self.SVoigt[0][0] + self.SVoigt[1][1] + self.SVoigt[2][2]) / 3
        b = (self.SVoigt[1][2] + self.SVoigt[0][2] + self.SVoigt[0][1]) / 3
        c = (self.SVoigt[3][3] + self.SVoigt[4][4] + self.SVoigt[5][5]) / 3

        KV = (A + 2 * B) / 3
        GV = (A - B + 3 * C) / 5

        KR = 1 / (3 * a + 6 * b)
        GR = 5 / (4 * a - 4 * b + 3 * c)

        KH = (KV + KR) / 2
        GH = (GV + GR) / 2

        return [
            [
                KV,
                1 / (1 / (3 * GV) + 1 / (9 * KV)),
                GV,
                (1 - 3 * GV / (3 * KV + GV)) / 2,
            ],
            [
                KR,
                1 / (1 / (3 * GR) + 1 / (9 * KR)),
                GR,
                (1 - 3 * GR / (3 * KR + GR)) / 2,
            ],
            [
                KH,
                1 / (1 / (3 * GH) + 1 / (9 * KH)),
                GH,
                (1 - 3 * GH / (3 * KH + GH)) / 2,
            ],
        ]

    def shear2D(self, x):
        ftol = 0.001
        xtol = 0.01

        def func1(z):
            return self.shear([x[0], x[1], z])

        r1 = optimize.minimize(
            func1,
            np.pi / 2.0,
            args=(),
            method="Powell",
            options={"xtol": xtol, "ftol": ftol},
        )  # , bounds=[(0.0,np.pi)])

        def func2(z):
            return -self.shear([x[0], x[1], z])

        r2 = optimize.minimize(
            func2,
            np.pi / 2.0,
            args=(),
            method="Powell",
            options={"xtol": xtol, "ftol": ftol},
        )  # , bounds=[(0.0,np.pi)])
        return (float(r1.fun), -float(r2.fun))

    def shear3D(self, x, y, guess1=np.pi / 2.0, guess2=np.pi / 2.0):
        tol = 0.005

        def func1(z):
            return self.shear([x, y, z])

        r1 = optimize.minimize(
            func1, guess1, args=(), method="COBYLA", options={"tol": tol}
        )  # , bounds=[(0.0,np.pi)])

        def func2(z):
            return -self.shear([x, y, z])

        r2 = optimize.minimize(
            func2, guess2, args=(), method="COBYLA", options={"tol": tol}
        )  # , bounds=[(0.0,np.pi)])
        return (float(r1.fun), -float(r2.fun), float(r1.x), float(r2.x))

    def Poisson2D(self, x):
        ftol = 0.001
        xtol = 0.01

        def func1(z):
            return self.Poisson([x[0], x[1], z])

        r1 = optimize.minimize(
            func1,
            np.pi / 2.0,
            args=(),
            method="Powell",
            options={"xtol": xtol, "ftol": ftol},
        )  # , bounds=[(0.0,np.pi)])

        def func2(z):
            return -self.Poisson([x[0], x[1], z])

        r2 = optimize.minimize(
            func2,
            np.pi / 2.0,
            args=(),
            method="Powell",
            options={"xtol": xtol, "ftol": ftol},
        )  # , bounds=[(0.0,np.pi)])
        return (min(0, float(r1.fun)), max(0, float(r1.fun)), -float(r2.fun))

    def poisson3D(self, x, y, guess1=np.pi / 2.0, guess2=np.pi / 2.0):
        tol = 0.005

        def func1(z):
            return self.Poisson([x, y, z])

        r1 = optimize.minimize(
            func1, guess1, args=(), method="COBYLA", options={"tol": tol}
        )  # , bounds=[(0.0,np.pi)])

        def func2(z):
            return -self.Poisson([x, y, z])

        r2 = optimize.minimize(
            func2, guess2, args=(), method="COBYLA", options={"tol": tol}
        )  # , bounds=[(0.0,np.pi)])
        return (
            min(0, float(r1.fun)),
            max(0, float(r1.fun)),
            -float(r2.fun),
            float(r1.x),
            float(r2.x),
        )


    def compressionSpeed2D(self, x, density = None):
        ftol = 0.001
        xtol = 0.01

        def func1(z):
            return (3*self.averages()[2][0]*(10**9)*(1-self.Poisson([x[0], x[1], z]))/(2*density*(1+self.Poisson([x[0], x[1], z]))))**0.5


        r1 = optimize.minimize(
            func1,
            np.pi / 2.0,
            args=(),
            method="Powell",
            options={"xtol": xtol, "ftol": ftol},
        )  # , bounds=[(0.0,np.pi)])

        def func2(z):
            return -(3*self.averages()[2][0]*(10**9)*(1-self.Poisson([x[0], x[1], z]))/(2*density*(1+self.Poisson([x[0], x[1], z]))))**0.5


        r2 = optimize.minimize(
            func2,
            np.pi / 2.0,
            args=(),
            method="Powell",
            options={"xtol": xtol, "ftol": ftol},
        )  # , bounds=[(0.0,np.pi)])
        return (float(r1.fun), -float(r2.fun))


    def compressionSpeed3D(self, x, y, guess1=np.pi / 2.0, guess2=np.pi / 2.0, density = None):
        tol = 0.005

        def func1(z):
            return (3*self.averages()[2][0]*(10**9)*(1-self.Poisson([x, y, z]))/(2*density*(1+self.Poisson([x, y, z]))))**0.5

        r1 = optimize.minimize(
            func1, guess1, args=(), method="COBYLA", options={"tol": tol}
        )  # , bounds=[(0.0,np.pi)])

        def func2(z):
            return -(3*self.averages()[2][0]*(10**9)*(1-self.Poisson([x, y, z]))/(2*density*(1+self.Poisson([x, y, z]))))**0.5

        r2 = optimize.minimize(
            func2, guess2, args=(), method="COBYLA", options={"tol": tol}
        )  # , bounds=[(0.0,np.pi)])
        return (float(r1.fun), -float(r2.fun), float(r1.x), float(r2.x))

    def shearSpeed2D(self, x, density = None):
        ftol = 0.001
        xtol = 0.01

        def func1(z):
            return ((1/density)*(10**9)*self.shear([x[0], x[1], z]))**0.5
        r1 = optimize.minimize(
            func1,
            np.pi / 2.0,
            args=(),
            method="Powell",
            options={"xtol": xtol, "ftol": ftol},
        )  # , bounds=[(0.0,np.pi)])

        def func2(z):
            return -((1/density)*(10**9)*self.shear([x[0], x[1], z]))**0.5

        r2 = optimize.minimize(
            func2,
            np.pi / 2.0,
            args=(),
            method="Powell",
            options={"xtol": xtol, "ftol": ftol},
        )  # , bounds=[(0.0,np.pi)])
        return (float(r1.fun), -float(r2.fun))


    def shearSpeed3D(self, x, y, guess1=np.pi / 2.0, guess2=np.pi / 2.0, density=None):
        tol = 0.005

        def func1(z):
            return ((1/density)*(10**9)*self.shear([x, y, z]))**0.5

        r1 = optimize.minimize(
            func1, guess1, args=(), method="COBYLA", options={"tol": tol}
        )  # , bounds=[(0.0,np.pi)])

        def func2(z):
            return ((1/density)*(10**9)*self.shear([x, y, z]))**0.5

        r2 = optimize.minimize(
            func2, guess2, args=(), method="COBYLA", options={"tol": tol}
        )  # , bounds=[(0.0,np.pi)])
        return (float(r1.fun), -float(r2.fun), float(r1.x), float(r2.x))


    def ratio_compressional_shear2D(self, x, density = None):
        ftol = 0.001
        xtol = 0.01

        def v_p(z):
            return (3*self.averages()[2][0]*(10**9)*(1-self.Poisson([x[0], x[1], z]))/(2*density*(1+self.Poisson([x[0], x[1], z]))))**0.5

        def v_s(z):
            return ((1/density)*(10**9)*self.shear([x[0], x[1], z]))**0.5


        def func1(z):
            return (v_p(z) /v_s(z))**2

        r1 = optimize.minimize(
            func1,
            np.pi / 2.0,
            args=(),
            method="Powell",
            options={"xtol": xtol, "ftol": ftol},
        )  # , bounds=[(0.0,np.pi)])

        def func2(z):
            return -(v_p(z) /v_s(z))**2

        r2 = optimize.minimize(
            func2,
            np.pi / 2.0,
            args=(),
            method="Powell",
            options={"xtol": xtol, "ftol": ftol},
        )  # , bounds=[(0.0,np.pi)])
        return (float(r1.fun), -float(r2.fun))


    def ratio_compressional_shear3D(self, x, y, guess1=np.pi / 2.0, guess2=np.pi / 2.0,density = None):
        tol = 0.005


        def v_p(z):
            return (3*self.averages()[2][0]*(10**9)*(1-self.Poisson([x, y, z]))/(2*density*(1+self.Poisson([x, y, z]))))**0.5

        def v_s(z):
            return ((1/density)*(10**9)*self.shear([x, y, z]))**0.5


        def func1(z):
            return (v_p(z) /v_s(z))**2

        r1 = optimize.minimize(
            func1, guess1, args=(), method="COBYLA", options={"tol": tol}
        )  # , bounds=[(0.0,np.pi)])

        def func2(z):
            return -(v_p(z) /v_s(z))**2

        r2 = optimize.minimize(
            func2, guess2, args=(), method="COBYLA", options={"tol": tol}
        )  # , bounds=[(0.0,np.pi)])
        return (float(r1.fun), -float(r2.fun), float(r1.x), float(r2.x))


    def debyeSpeed2D(self, x,density = None):
        ftol = 0.001
        xtol = 0.01

        def v_p(z):
            return (3*self.averages()[2][0]*(10**9)*(1-self.Poisson([x[0], x[1], z]))/(2*density*(1+self.Poisson([x[0], x[1], z]))))**0.5

        def v_s(z):
            return ((1/density)*(10**9)*self.shear([x[0], x[1], z]))**0.5

        def func1(z):
            return v_p(z)*v_s(z)/(2*v_s(z)**3 +v_p(z)**3)**(1/3)
        r1 = optimize.minimize(
            func1,
            np.pi / 2.0,
            args=(),
            method="Powell",
            options={"xtol": xtol, "ftol": ftol},
        )  # , bounds=[(0.0,np.pi)])

        def func2(z):
            return -v_p(z)*v_s(z)/(2*v_s(z)**3 +v_p(z)**3)**(1/3)

        r2 = optimize.minimize(
            func2,
            np.pi / 2.0,
            args=(),
            method="Powell",
            options={"xtol": xtol, "ftol": ftol},
        )  # , bounds=[(0.0,np.pi)])
        return (float(r1.fun), -float(r2.fun))


    def debyeSpeed3D(self, x, y, guess1=np.pi / 2.0, guess2=np.pi / 2.0,density= None):
        tol = 0.005

        def v_p(z):
            return (3*self.averages()[2][0]*(10**9)*(1-self.Poisson([x, y, z]))/(2*density*(1+self.Poisson([x, y, z]))))**0.5

        def v_s(z):
            return ((1/density)*(10**9)*self.shear([x, y, z]))**0.5

        def func1(z):
            return v_p(z)*v_s(z)/(2*v_s(z)**3 +v_p(z)**3)**(1/3)

        r1 = optimize.minimize(
            func1, guess1, args=(), method="COBYLA", options={"tol": tol}
        )  # , bounds=[(0.0,np.pi)])

        def func2(z):
            return -v_p(z)*v_s(z)/(2*v_s(z)**3 +v_p(z)**3)**(1/3)

        r2 = optimize.minimize(
            func2, guess2, args=(), method="COBYLA", options={"tol": tol}
        )  # , bounds=[(0.0,np.pi)])
        return (float(r1.fun), -float(r2.fun), float(r1.x), float(r2.x))


class ElasticOrtho(Elastic):
    """An elastic tensor, for the specific case of an orthorhombic system"""

    def __init__(self, arg):
        """Initialize from a matrix, or from an Elastic object"""
        if type(arg) == str:
            Elastic.__init__(self, arg)
        elif isinstance(arg, Elastic):
            self.CVoigt = arg.CVoigt
            self.SVoigt = arg.SVoigt
            self.Smat = arg.Smat
        else:
            raise TypeError(
                "ElasticOrtho constructor argument should be string or Elastic object"
            )

    def Young(self, x):
        ct2 = math.cos(x[0]) ** 2
        st2 = 1 - ct2
        cf2 = math.cos(x[1]) ** 2
        sf2 = 1 - cf2
        s11 = self.Smat[0][0][0][0]
        s22 = self.Smat[1][1][1][1]
        s33 = self.Smat[2][2][2][2]
        s44 = 4 * self.Smat[1][2][1][2]
        s55 = 4 * self.Smat[0][2][0][2]
        s66 = 4 * self.Smat[0][1][0][1]
        s12 = self.Smat[0][0][1][1]
        s13 = self.Smat[0][0][2][2]
        s23 = self.Smat[1][1][2][2]
        return 1 / (
            ct2 ** 2 * s33
            + 2 * cf2 * ct2 * s13 * st2
            + cf2 * ct2 * s55 * st2
            + 2 * ct2 * s23 * sf2 * st2
            + ct2 * s44 * sf2 * st2
            + cf2 ** 2 * s11 * st2 ** 2
            + 2 * cf2 * s12 * sf2 * st2 ** 2
            + cf2 * s66 * sf2 * st2 ** 2
            + s22 * sf2 ** 2 * st2 ** 2
        )

    def LC(self, x):
        ct2 = math.cos(x[0]) ** 2
        cf2 = math.cos(x[1]) ** 2
        s11 = self.Smat[0][0][0][0]
        s22 = self.Smat[1][1][1][1]
        s33 = self.Smat[2][2][2][2]
        s12 = self.Smat[0][0][1][1]
        s13 = self.Smat[0][0][2][2]
        s23 = self.Smat[1][1][2][2]
        return 1000 * (
            ct2 * (s13 + s23 + s33)
            + (cf2 * (s11 + s12 + s13) + (s12 + s22 + s23) * (1 - cf2)) * (1 - ct2)
        )

    def shear(self, x):
        ct = math.cos(x[0])
        ct2 = ct * ct
        st2 = 1 - ct2
        cf = math.cos(x[1])
        sf = math.sin(x[1])
        sf2 = sf * sf
        cx = math.cos(x[2])
        cx2 = cx * cx
        sx = math.sin(x[2])
        sx2 = 1 - cx2
        s11 = self.Smat[0][0][0][0]
        s22 = self.Smat[1][1][1][1]
        s33 = self.Smat[2][2][2][2]
        s44 = 4 * self.Smat[1][2][1][2]
        s55 = 4 * self.Smat[0][2][0][2]
        s66 = 4 * self.Smat[0][1][0][1]
        s12 = self.Smat[0][0][1][1]
        s13 = self.Smat[0][0][2][2]
        s23 = self.Smat[1][1][2][2]
        r = (
            ct2 * ct2 * cx2 * s44 * sf2
            + cx2 * s44 * sf2 * st2 * st2
            + 4 * cf ** 3 * ct * cx * (-2 * s11 + 2 * s12 + s66) * sf * st2 * sx
            + 2
            * cf
            * ct
            * cx
            * sf
            * (
                ct2 * (s44 - s55)
                + (
                    4 * s13
                    - 4 * s23
                    - s44
                    + s55
                    - 4 * s12 * sf2
                    + 4 * s22 * sf2
                    - 2 * s66 * sf2
                )
                * st2
            )
            * sx
            + s66 * sf2 * sf2 * st2 * sx2
            + cf ** 4 * st2 * (4 * ct2 * cx2 * s11 + s66 * sx2)
            + ct2
            * (
                2 * cx2 * (2 * s33 + sf2 * (-4 * s23 - s44 + 2 * s22 * sf2)) * st2
                + s55 * sf2 * sx2
            )
            + cf ** 2
            * (
                ct2 * ct2 * cx2 * s55
                + ct2
                * (
                    -2 * cx2 * (4 * s13 + s55 - 2 * (2 * s12 + s66) * sf2) * st2
                    + s44 * sx2
                )
                + st2
                * (
                    cx2 * s55 * st2
                    + 2 * (2 * s11 - 4 * s12 + 2 * s22 - s66) * sf2 * sx2
                )
            )
        )
        return 1 / r

    def Poisson(self, x):
        ct = math.cos(x[0])
        ct2 = ct * ct
        st2 = 1 - ct2
        cf = math.cos(x[1])
        sf = math.sin(x[1])
        cx = math.cos(x[2])
        sx = math.sin(x[2])
        s11 = self.Smat[0][0][0][0]
        s22 = self.Smat[1][1][1][1]
        s33 = self.Smat[2][2][2][2]
        s44 = 4 * self.Smat[1][2][1][2]
        s55 = 4 * self.Smat[0][2][0][2]
        s66 = 4 * self.Smat[0][1][0][1]
        s12 = self.Smat[0][0][1][1]
        s13 = self.Smat[0][0][2][2]
        s23 = self.Smat[1][1][2][2]

        return (
            -(ct ** 2 * cx ** 2 * s33 * st2)
            - cf ** 2 * cx ** 2 * s13 * st2 * st2
            - cx ** 2 * s23 * sf ** 2 * st2 * st2
            + ct * cx * s44 * sf * st2 * (ct * cx * sf + cf * sx)
            - ct ** 2 * s23 * (ct * cx * sf + cf * sx) ** 2
            - cf ** 2 * s12 * st2 * (ct * cx * sf + cf * sx) ** 2
            - s22 * sf ** 2 * st2 * (ct * cx * sf + cf * sx) ** 2
            + cf * ct * cx * s55 * st2 * (cf * ct * cx - sf * sx)
            - cf * s66 * sf * st2 * (ct * cx * sf + cf * sx) * (cf * ct * cx - sf * sx)
            - ct ** 2 * s13 * (cf * ct * cx - sf * sx) ** 2
            - cf ** 2 * s11 * st2 * (cf * ct * cx - sf * sx) ** 2
            - s12 * sf ** 2 * st2 * (cf * ct * cx - sf * sx) ** 2
        ) / (
            ct ** 4 * s33
            + 2 * cf ** 2 * ct ** 2 * s13 * st2
            + cf ** 2 * ct ** 2 * s55 * st2
            + 2 * ct ** 2 * s23 * sf ** 2 * st2
            + ct ** 2 * s44 * sf ** 2 * st2
            + cf ** 4 * s11 * st2 * st2
            + 2 * cf ** 2 * s12 * sf ** 2 * st2 * st2
            + cf ** 2 * s66 * sf ** 2 * st2 * st2
            + s22 * sf ** 4 * st2 * st2
        )


#########################################################################################
# Data genertaation
################################################################################################
# ELATE : basic usage of the tool, only 2D plots
# YOUNG3D : visualize Young's modulus in 3D
# LC3D : visualize Linear compressiblity in 3D
# SHEAR3D : visualize Shear modulus in 3D
# POISSON3D : visualize Poisson ratio in 3D
################################################################################################




from distutils.core import Extension, setup
from setuptools import find_packages
import json
import pathlib
import os


def get_version_info():
    """Retrieve version info from setup.json
    Method adopted from PyChemia."""

    basepath = pathlib.Path(__file__).parent.absolute()
    rf = open(str(basepath) + os.sep + "setup.json")
    release_data = json.load(rf)
    rf.close()

    return release_data


def write_version_py(filename="mechelastic/version.py"):
    """Write mechelastic/version.py. Adopted from
    PyChemia."""

    versioninfo_string = """
# THIS FILE IS GENERATED FROM MECHELASTIC SETUP.PY.
name = '%(name)s'
version = '%(version)s'
description = '%(description)s'
url = '%(url)s'
author = '%(author)s'
email = '%(email)s'
status = '%(status)s'
copyright = '%(copyright)s'
date = '%(date)s'
"""
    release_data = get_version_info()

    a = open(filename, "w")
    try:
        a.write(
            versioninfo_string
            % {
                "name": release_data["name"],
                "version": release_data["version"],
                "description": release_data["description"],
                "url": release_data["url"],
                "author": release_data["author"],
                "email": release_data["email"],
                "status": release_data["status"],
                "copyright": release_data["copyright"],
                "date": release_data["date"],
            }
        )
    finally:
        a.close()
    return release_data


data = write_version_py()

setup(
    name=data["name"],
    version=data["version"],
    description=data["description"],
    author=data["author"],
    author_email=data["email"],
    url=data["url"],
    download_url=data["download_url"],
    license="LICENSE.txt",
    install_requires=[
        "numpy",
        "spglib",
        "pathlib",
        "pyvista",
        "intersect",
        "networkx",
        "dicttoxml",
    ],
    data_files=[("", ["LICENSE.txt"])],
    package_data={"": ["setup.json"]},
    scripts=["scripts/MechElastic.py"],
    packages=find_packages(exclude=["scripts", "examples"]),
)



#!/usr/bin/env python


from .comms import printer
from .parsers import VaspOutcar
from .parsers import AbinitOutput
from .parsers import QE_ElaStic_Parser
from .parsers import QE_thermo_pw_Parser
from .core import ElasticProperties
from .core import ElasticProperties2D


def calculate_elastic(
    infile="OUTCAR",
    dim="3D",
    elastic_tensor=None,
    crystal=None,
    lattice_type=None,
    code="vasp",
    anaddbfile=None,
    qe_outfile=None,
    adjust_pressure=True,
    verbose=True,
    outfile="elastic_properties.txt",
):
    """
    Calculate the elastic properties of a material from a DFT calculation.

    Parameters
    ----------
    infile : str, optional
        Path to the input file which is a DFT calculation outputfile. The default is "OUTCAR".
    dim : str, optional
        Dimension of the structure. The default is "3D".
    crystal : str, optional
        Crystal family (only used in 3D). The default is None.
    elastic_tensor : float, optional
        The elastic tensor. This option is useful if one does not want to use a DFT output.
        The default is None.
    lattice_type : TYPE, optional
        2D lattice type. The default is None.
    code : str, optional
        DFT code used to generate the outputs. The default is "vasp".
    anaddbfile : str, optional
        Path to the DDB file (applicable only in abinit). The default is None.
    qe_outfile : str, optional
        Path to the Quantum Espresso output file. The default is None.
    adjust_pressure : bool, optional
        To adjust the cell pressure according to the output file. The default is True.
    verbose : str, optional
        To print the progress of the elastic calculations. The default is True.
    outfile : TYPE, optional
        Path to the desired output file. Acceptable files are JSON, XML, TXT.
        The default is "elastic_properties.txt".

    Returns
    -------
    elastic_properties : TYPE
        DESCRIPTION.

    """

    # welcome message
    if verbose:
        printer.print_mechelastic()
        if code != None:
            print("\nThis matrix was computed from " + code)
    elastic_tensor = elastic_tensor
    structure = None
    lattice_constant = None
    crystal_type = crystal

    # calling parser
    if code == "vasp" and elastic_tensor is None:
        output = VaspOutcar(
            infile=infile, adjust_pressure=adjust_pressure, verbose=verbose
        )
        elastic_tensor = output.elastic_tensor
        structure = output.structure
        lattice_constant = output.lattice_constant

    elif code == "abinit":
        output = AbinitOutput(infile=infile, anaddbfile=anaddbfile)
        elastic_tensor = output.elastic_tensor
        structure = output.structure
        lattice_constant = output.lattice_constant

    elif code == "qe_ElaStic":
        output = QE_ElaStic_Parser(outfile=qe_outfile, infile=infile)
        elastic_tensor = output.elastic_tensor
        structure = output.structure
        lattice_constant = output.lattice_constant

    elif code == "qe_thermo_pw":
        output = QE_thermo_pw_Parser(outfile=qe_outfile, infile=infile)
        elastic_tensor = output.elastic_tensor
        structure = output.structure
        lattice_constant = output.lattice_constant

    # elastic constants calculation for 3D materials
    if dim == "3D":
        elastic_properties = ElasticProperties(
            elastic_tensor, structure, crystal_type, verbose=verbose
        )
        if verbose:
            print(elastic_properties)
        if "json" in outfile:
            elastic_properties.to_json(outfile)
        elif "xml" in outfile:
            elastic_properties.to_xml(outfile)
        else:
            elastic_properties.to_file(outfile)

    # elastic constants calculation for 2D materials
    elif dim == "2D":
        elastic_properties = ElasticProperties2D(
            elastic_tensor, lattice_constant, lattice_type=lattice_type
        )
        elastic_properties.print_properties()

    if verbose:
        print("\nThanks! See you later. ")
    return elastic_properties
